"""
Additive model for predicting gene expression from fragments
"""


import torch
import torch_scatter
import math
import numpy as np
import xarray as xr
import pandas as pd

import pickle

from chromatinhd.embedding import EmbeddingTensor
from chromatinhd.models import HybridModel
from chromatinhd.flow import Linked, Flow, Stored

from chromatinhd.models.pred.loader.minibatches import Minibatcher
from chromatinhd.models.pred.loader.transcriptome_fragments import (
    TranscriptomeFragments,
)
from chromatinhd.models.pred.trainer import Trainer
from chromatinhd.loaders import LoaderPool2
from chromatinhd.optim import SparseDenseAdam


def paircor(x, y, dim=0, eps=0.1):
    divisor = (y.std(dim) * x.std(dim)) + eps
    cor = ((x - x.mean(dim, keepdims=True)) * (y - y.mean(dim, keepdims=True))).mean(
        dim
    ) / divisor
    return cor


def paircor_loss(x, y):
    return -paircor(x, y).mean() * 100


def gene_paircor_loss(x, y):
    return -paircor(x, y) * 100


class SineEncoding(torch.nn.Module):
    def __init__(self, n_frequencies):
        super().__init__()

        self.register_buffer(
            "frequencies",
            torch.tensor(
                [
                    [1 / 1000 ** (2 * i / n_frequencies)] * 2
                    for i in range(1, n_frequencies + 1)
                ]
            ).flatten(-2),
        )
        self.register_buffer(
            "shifts",
            torch.tensor(
                [[0, torch.pi / 2] for _ in range(1, n_frequencies + 1)]
            ).flatten(-2),
        )

        self.n_embedding_dimensions = n_frequencies * 2 * 2

    def forward(self, coordinates):
        embedding = torch.sin(
            (coordinates[..., None] * self.frequencies + self.shifts).flatten(-2)
        )
        return embedding


class FragmentEmbedder(torch.nn.Module):
    dropout_rate = 0.0

    def __init__(
        self,
        n_genes,
        n_frequencies=10,
        n_embedding_dimensions=5,
        nonlinear=True,
        dropout_rate=0.0,
        **kwargs
    ):

        self.n_embedding_dimensions = n_embedding_dimensions

        self.nonlinear = nonlinear
        self.dropout_rate = dropout_rate

        super().__init__(**kwargs)

        self.sine_encoding = SineEncoding(n_frequencies=n_frequencies)

        # default initialization same as a torch.nn.Linear
        self.bias1 = EmbeddingTensor(
            n_genes,
            (self.n_embedding_dimensions,),
            sparse=True,
        )
        self.bias1.data.zero_()

        self.weight1 = EmbeddingTensor(
            n_genes,
            (
                self.sine_encoding.n_embedding_dimensions,
                self.n_embedding_dimensions,
            ),
            sparse=True,
        )
        stdv = 1.0 / math.sqrt(self.weight1.shape[-1])  # / 100
        self.weight1.data.uniform_(-stdv, stdv)

    def forward(self, coordinates, gene_ix):
        embedding = self.sine_encoding(coordinates)
        embedding = torch.einsum(
            "ab,abc->ac", embedding, self.weight1(gene_ix)
        ) + self.bias1(gene_ix)
        # embedding = (embedding[..., None] * self.weight1[gene_ix]).sum(-2)

        # non-linear
        if self.nonlinear is True:
            embedding = torch.sigmoid(embedding)
        elif isinstance(self.nonlinear, str) and self.nonlinear == "relu":
            embedding = torch.relu(embedding)
        elif isinstance(self.nonlinear, str) and self.nonlinear == "elu":
            embedding = torch.nn.functional.elu(embedding)

        if self.dropout_rate > 0:
            embedding = torch.nn.functional.dropout(embedding, self.dropout_rate)

        return embedding

    def parameters_sparse(self):
        return [self.bias1.weight, self.weight1.weight]


class FragmentEmbedderCounter(torch.nn.Sequential):
    """
    Dummy embedding of fragments in a single embedding dimension of ones
    """

    def __init__(self, *args, **kwargs):
        self.n_embedding_dimensions = 1
        super().__init__(*args, **kwargs)

    def forward(self, coordinates, gene_ix):
        return torch.ones(
            (*coordinates.shape[:-1], 1), device=coordinates.device, dtype=torch.float
        )


class EmbeddingGenePooler(torch.nn.Module):
    """
    Pools fragments across genes and cells
    """

    def __init__(self, reduce="sum"):
        self.reduce = reduce
        super().__init__()

    def forward(self, embedding, fragment_cellxgene_ix, cell_n, gene_n):
        if self.reduce == "mean":
            cellxgene_embedding = torch_scatter.segment_mean_coo(
                embedding, fragment_cellxgene_ix, dim_size=cell_n * gene_n
            )
        elif self.reduce == "sum":
            cellxgene_embedding = torch_scatter.segment_sum_coo(
                embedding, fragment_cellxgene_ix, dim_size=cell_n * gene_n
            )
        else:
            raise ValueError()
        cell_gene_embedding = cellxgene_embedding.reshape(
            (cell_n, gene_n, cellxgene_embedding.shape[-1])
        )
        return cell_gene_embedding


class EmbeddingToExpression(torch.nn.Module):
    """
    Predicts gene expression using a [cell, gene, component] embedding in a gene-specific manner
    """

    def __init__(
        self, n_genes, n_embedding_dimensions=5, initialization="default", **kwargs
    ):
        self.n_genes = n_genes
        self.n_embedding_dimensions = n_embedding_dimensions

        super().__init__()

        # set bias to empirical mean
        self.bias1 = EmbeddingTensor(
            n_genes,
            tuple(),
            sparse=True,
        )
        self.bias1.data[:] = 0.0

        self.weight1 = EmbeddingTensor(
            n_genes,
            (n_embedding_dimensions,),
            sparse=True,
        )
        if initialization == "ones":
            self.weight1.data[:] = 1.0
        elif initialization == "default":
            self.weight1.data[:, :5] = 1.0
            self.weight1.data[:, 5:] = 0.0
            self.weight1.data[:, -1] = -1.0
        elif initialization == "smaller":
            stdv = 1.0 / math.sqrt(self.weight1.data.size(-1)) / 100
            self.weight1.data.uniform_(-stdv, stdv)

    def forward(self, cell_gene_embedding, gene_ix):
        out = (cell_gene_embedding * self.weight1(gene_ix)).sum(-1) + self.bias1(
            gene_ix
        ).squeeze()
        return out

    def parameters_sparse(self):
        return [self.bias1.weight, self.weight1.weight]


class Model(torch.nn.Module, HybridModel):
    def __init__(
        self,
        n_genes,
        dummy=False,
        n_frequencies=50,
        reduce="sum",
        nonlinear=True,
        n_embedding_dimensions=10,
        dropout_rate=0.0,
        embedding_to_expression_initialization="default",
        **kwargs
    ):
        super().__init__()

        if dummy:
            self.fragment_embedder = FragmentEmbedderCounter()
        else:
            self.fragment_embedder = FragmentEmbedder(
                n_frequencies=n_frequencies,
                n_genes=n_genes,
                nonlinear=nonlinear,
                n_embedding_dimensions=n_embedding_dimensions,
                dropout_rate=dropout_rate,
            )
        self.embedding_gene_pooler = EmbeddingGenePooler(reduce=reduce)
        self.embedding_to_expression = EmbeddingToExpression(
            n_genes=n_genes,
            n_embedding_dimensions=self.fragment_embedder.n_embedding_dimensions,
            initialization=embedding_to_expression_initialization,
        )

    def forward(self, data):
        fragment_embedding = self.fragment_embedder(
            data.fragments.coordinates, data.fragments.genemapping
        )
        cell_gene_embedding = self.embedding_gene_pooler(
            fragment_embedding,
            data.fragments.local_cellxgene_ix,
            data.minibatch.n_cells,
            data.minibatch.n_genes,
        )
        expression_predicted = self.embedding_to_expression(
            cell_gene_embedding, data.minibatch.genes_oi_torch
        )
        return expression_predicted

    def forward_loss(self, data):
        expression_predicted = self.forward(data)
        expression_true = data.transcriptome.value
        return paircor_loss(expression_predicted, expression_true)

    def forward_gene_loss(self, data):
        expression_predicted = self.forward(data)
        expression_true = data.transcriptome.value
        return gene_paircor_loss(expression_predicted, expression_true)

    def forward_multiple(self, data, fragments_oi, min_fragments=1):
        fragment_embedding = self.fragment_embedder(
            data.fragments.coordinates, data.fragments.genemapping
        )

        total_n_fragments = torch.bincount(
            data.fragments.local_cellxgene_ix,
            minlength=data.minibatch.n_genes * data.minibatch.n_cells,
        ).reshape((data.minibatch.n_cells, data.minibatch.n_genes))

        total_cell_gene_embedding = self.embedding_gene_pooler.forward(
            fragment_embedding,
            data.fragments.local_cellxgene_ix,
            data.minibatch.n_cells,
            data.minibatch.n_genes,
        )

        total_expression_predicted = self.embedding_to_expression.forward(
            total_cell_gene_embedding, data.minibatch.genes_oi_torch
        )

        for fragments_oi_ in fragments_oi:
            if (fragments_oi_ is not None) and ((~fragments_oi_).sum() > min_fragments):
                lost_fragments_oi = ~fragments_oi_
                lost_local_cellxgene_ix = data.fragments.local_cellxgene_ix[
                    lost_fragments_oi
                ]
                n_fragments = total_n_fragments - torch.bincount(
                    lost_local_cellxgene_ix,
                    minlength=data.minibatch.n_genes * data.minibatch.n_cells,
                ).reshape((data.minibatch.n_cells, data.minibatch.n_genes))
                cell_gene_embedding = (
                    total_cell_gene_embedding
                    - self.embedding_gene_pooler.forward(
                        fragment_embedding[lost_fragments_oi],
                        lost_local_cellxgene_ix,
                        data.minibatch.n_cells,
                        data.minibatch.n_genes,
                    )
                )

                expression_predicted = self.embedding_to_expression.forward(
                    cell_gene_embedding, data.minibatch.genes_oi_torch
                )
            else:
                n_fragments = total_n_fragments
                expression_predicted = total_expression_predicted

            yield expression_predicted, n_fragments

    def train_model(self, fragments, transcriptome, fold, device="cuda"):
        # set up minibatchers and loaders
        minibatcher_train = Minibatcher(
            fold["cells_train"],
            range(fragments.n_genes),
            n_genes_step=500,
            n_cells_step=200,
        )
        minibatcher_validation = Minibatcher(
            fold["cells_validation"],
            range(fragments.n_genes),
            n_genes_step=10,
            n_cells_step=10000,
            permute_cells=False,
            permute_genes=False,
        )

        loaders_train = LoaderPool2(
            TranscriptomeFragments,
            dict(
                transcriptome=transcriptome,
                fragments=fragments,
                cellxgene_batch_size=minibatcher_train.cellxgene_batch_size,
            ),
            n_workers=10,
        )
        loaders_validation = LoaderPool2(
            TranscriptomeFragments,
            dict(
                transcriptome=transcriptome,
                fragments=fragments,
                cellxgene_batch_size=minibatcher_validation.cellxgene_batch_size,
            ),
            n_workers=5,
        )

        trainer = Trainer(
            self,
            loaders_train,
            loaders_validation,
            minibatcher_train,
            minibatcher_validation,
            SparseDenseAdam(
                self.parameters_sparse(),
                self.parameters_dense(),
                lr=1e-2,
                weight_decay=1e-5,
            ),
            n_epochs=30,
            checkpoint_every_epoch=1,
            optimize_every_step=1,
            device=device,
        )

        trainer.train()
        # trainer.trace.plot()

    def get_prediction(
        self,
        fragments,
        transcriptome,
        cells=None,
        cell_ixs=None,
        genes=None,
        gene_ixs=None,
        device="cuda",
        return_raw=False,
    ):
        """
        Returns the prediction of a dataset
        """
        if cell_ixs is None:
            if cells is None:
                cells = fragments.obs.index
            fragments.obs["ix"] = np.arange(len(fragments.obs))
            cell_ixs = fragments.obs.loc[cells]["ix"].values
        if cells is None:
            cells = fragments.obs.index[cell_ixs]

        if gene_ixs is None:
            if genes is None:
                genes = fragments.var.index
            fragments.var["ix"] = np.arange(len(fragments.var))
            gene_ixs = fragments.var.loc[genes]["ix"].values
        if genes is None:
            genes = fragments.var.index[gene_ixs]

        minibatches = Minibatcher(
            cell_ixs,
            gene_ixs,
            n_genes_step=500,
            n_cells_step=200,
            use_all_cells=True,
            use_all_genes=True,
            permute_cells=False,
            permute_genes=False,
        )
        loaders = LoaderPool2(
            TranscriptomeFragments,
            dict(
                transcriptome=transcriptome,
                fragments=fragments,
                cellxgene_batch_size=minibatches.cellxgene_batch_size,
            ),
            n_workers=5,
        )
        loaders.initialize(minibatches)

        predicted = np.zeros((len(cell_ixs), fragments.n_genes))
        expected = np.zeros((len(cell_ixs), fragments.n_genes))
        n_fragments = np.zeros((len(cell_ixs), fragments.n_genes))

        cell_mapping = np.zeros(fragments.n_cells, dtype=np.int64)
        cell_mapping[cell_ixs] = np.arange(len(cell_ixs))

        device = "cuda"
        self.eval()
        self = self.to(device)

        for data in loaders:
            data = data.to(device)
            with torch.no_grad():
                pred_mb = self.forward(data)
            predicted[
                np.ix_(cell_mapping[data.minibatch.cells_oi], data.minibatch.genes_oi)
            ] = pred_mb.cpu().numpy()
            expected[
                np.ix_(cell_mapping[data.minibatch.cells_oi], data.minibatch.genes_oi)
            ] = data.transcriptome.value.cpu().numpy()
            n_fragments[
                np.ix_(cell_mapping[data.minibatch.cells_oi], data.minibatch.genes_oi)
            ] = (
                torch.bincount(
                    data.fragments.local_cellxgene_ix,
                    minlength=len(data.minibatch.cells_oi)
                    * len(data.minibatch.genes_oi),
                )
                .reshape(len(data.minibatch.cells_oi), len(data.minibatch.genes_oi))
                .cpu()
                .numpy()
            )

        self = self.to("cpu")

        if return_raw:
            return predicted, expected, n_fragments

        result = xr.Dataset(
            {
                "predicted": xr.DataArray(
                    predicted,
                    dims=("cell", "gene"),
                    coords={"cell": cells, "gene": fragments.var.index},
                ),
                "expected": xr.DataArray(
                    expected,
                    dims=("cell", "gene"),
                    coords={"cell": cells, "gene": fragments.var.index},
                ),
                "n_fragments": xr.DataArray(
                    n_fragments,
                    dims=("cell", "gene"),
                    coords={"cell": cells, "gene": fragments.var.index},
                ),
            }
        )
        return result

    def get_prediction_censored(
        self,
        fragments,
        transcriptome,
        censorer,
        cells=None,
        cell_ixs=None,
        genes=None,
        gene_ixs=None,
        device="cuda",
    ):
        """
        Returns the prediction of multiple censored dataset
        """
        if cell_ixs is None:
            if cells is None:
                cells = fragments.obs.index
            fragments.obs["ix"] = np.arange(len(fragments.obs))
            cell_ixs = fragments.obs.loc[cells]["ix"].values
        if cells is None:
            cells = fragments.obs.index[cell_ixs]

        if gene_ixs is None:
            if genes is None:
                genes = fragments.var.index
            fragments.var["ix"] = np.arange(len(fragments.var))
            gene_ixs = fragments.var.loc[genes]["ix"].values
        if genes is None:
            genes = fragments.var.index[gene_ixs]

        minibatcher = Minibatcher(
            cell_ixs,
            gene_ixs,
            n_genes_step=500,
            n_cells_step=5000,
            use_all_cells=True,
            use_all_genes=True,
            permute_cells=False,
            permute_genes=False,
        )
        loaders = LoaderPool2(
            TranscriptomeFragments,
            dict(
                transcriptome=transcriptome,
                fragments=fragments,
                cellxgene_batch_size=minibatcher.cellxgene_batch_size,
            ),
            n_workers=10,
        )
        loaders.initialize(minibatcher)

        predicted = np.zeros((len(censorer), len(cell_ixs), len(gene_ixs)), dtype=float)
        expected = np.zeros((len(cell_ixs), len(gene_ixs)), dtype=float)
        n_fragments = np.zeros((len(censorer), len(cell_ixs), len(gene_ixs)), dtype=int)

        cell_mapping = np.zeros(fragments.n_cells, dtype=np.int64)
        cell_mapping[cell_ixs] = np.arange(len(cell_ixs))
        gene_mapping = np.zeros(fragments.n_genes, dtype=np.int64)
        gene_mapping[gene_ixs] = np.arange(len(gene_ixs))

        self.eval()
        self.to(device)
        for data in loaders:
            data = data.to(device)
            fragments_oi = censorer(data)

            with torch.no_grad():
                for design_ix, (
                    pred_mb,
                    n_fragments_oi_mb,
                ) in enumerate(self.forward_multiple(data, fragments_oi)):
                    ix = np.ix_(
                        [design_ix],
                        cell_mapping[data.minibatch.cells_oi],
                        gene_mapping[data.minibatch.genes_oi],
                    )
                    predicted[ix] = pred_mb.cpu().numpy()
                    n_fragments[ix] = n_fragments_oi_mb.cpu().numpy()
            expected[
                np.ix_(
                    cell_mapping[data.minibatch.cells_oi],
                    gene_mapping[data.minibatch.genes_oi],
                )
            ] = data.transcriptome.value.cpu().numpy()

        self.to("cpu")

        return predicted, expected, n_fragments


class Models(Flow):
    n_models = Stored("n_models")

    @property
    def models_path(self):
        path = self.path / "models"
        path.mkdir(exist_ok=True)
        return path

    def train_models(self, fragments, transcriptome, folds, device="cuda"):
        self.n_models = len(folds)
        for fold_ix, fold in [(fold_ix, fold) for fold_ix, fold in enumerate(folds)]:
            desired_outputs = [self.models_path / ("model_" + str(fold_ix) + ".pkl")]
            force = False
            if not all([desired_output.exists() for desired_output in desired_outputs]):
                force = True

            if force:
                model = Model(
                    n_genes=fragments.n_genes,
                )
                model.train_model(fragments, transcriptome, fold, device=device)

                model = model.to("cpu")

                pickle.dump(
                    model,
                    open(self.models_path / ("model_" + str(fold_ix) + ".pkl"), "wb"),
                )

    def __getitem__(self, ix):
        return pickle.load(
            (self.models_path / ("model_" + str(ix) + ".pkl")).open("rb")
        )

    def __len__(self):
        return self.n_models

    def __iter__(self):
        for ix in range(len(self)):
            yield self[ix]

    def get_gene_cors(self, fragments, transcriptome, folds, device="cuda"):
        cor_predicted = np.zeros((len(fragments.var.index), len(folds)))
        cor_n_fragments = np.zeros((len(fragments.var.index), len(folds)))
        n_fragments = np.zeros((len(fragments.var.index), len(folds)))
        for model_ix, (model, fold) in enumerate(zip(self, folds)):
            prediction = model.get_prediction(
                fragments, transcriptome, cell_ixs=fold["cells_test"], device=device
            )

            cor_predicted[:, model_ix] = paircor(
                prediction["predicted"].values, prediction["expected"].values
            )
            cor_n_fragments[:, model_ix] = paircor(
                prediction["n_fragments"].values, prediction["expected"].values
            )

            n_fragments[:, model_ix] = prediction["n_fragments"].values.sum(0)
        cor_predicted = pd.Series(
            cor_predicted.mean(1), index=fragments.var.index, name="cor_predicted"
        )
        cor_n_fragments = pd.Series(
            cor_n_fragments.mean(1), index=fragments.var.index, name="cor_n_fragments"
        )
        n_fragments = pd.Series(
            n_fragments.mean(1), index=fragments.var.index, name="n_fragments"
        )
        result = pd.concat([cor_predicted, cor_n_fragments, n_fragments], axis=1)
        result["deltacor"] = result["cor_predicted"] - result["cor_n_fragments"]

        return result



import chromatinhd.data.clustering
import dataclasses
import torch


@dataclasses.dataclass
class Result:
    onehot: torch.Tensor

    def to(self, device):
        self.onehot = self.onehot.to(device)
        return self


class Clustering:
    """
    Provides clustering data for a minibatch.
    """

    def __init__(
        self,
        clustering: chromatinhd.data.clustering.Clustering,
    ):
        assert (
            clustering.labels.cat.categories == clustering.cluster_info.index
        ).all(), (
            clustering.labels.cat.categories,
            clustering.cluster_info.index,
        )
        self.onehot = torch.nn.functional.one_hot(
            torch.from_numpy(clustering.labels.cat.codes.values.copy()).to(torch.int64),
            clustering.n_clusters,
        )

    def load(self, minibatch):
        onehot = self.onehot[minibatch.cells_oi, :]
        return Result(onehot=onehot)



import numpy as np
import dataclasses
import itertools
import math
import torch


@dataclasses.dataclass
class Minibatch:
    cells_oi: np.ndarray
    genes_oi: np.ndarray
    phase: str = "train"
    device: str = "cpu"

    def items(self):
        return {"cells_oi": self.cells_oi, "genes_oi": self.genes_oi}

    def filter_genes(self, genes):
        genes_oi = self.genes_oi[genes[self.genes_oi]]

        return Minibatch(
            cells_oi=self.cells_oi,
            genes_oi=genes_oi,
            phase=self.phase,
        )

    def to(self, device):
        self.device = device

    @property
    def genes_oi_torch(self):
        return torch.from_numpy(self.genes_oi).to(self.device)

    @property
    def cells_oi_torch(self):
        return torch.from_numpy(self.genes_oi).to(self.device)

    @property
    def n_cells(self):
        return len(self.cells_oi)

    @property
    def n_genes(self):
        return len(self.genes_oi)


def create_bins_ordered(
    cells,
    genes,
    n_genes_total,
    n_genes_step=300,
    n_cells_step=1000,
    use_all=False,
    rg=None,
    permute_genes=True,
    permute_cells=True,
    **kwargs
):
    """
    Creates bins of cellxgene
    A number of cell and gene bins are created first, and then we create the product between the two
    """
    if rg is None:
        rg = np.random.RandomState()
    if permute_cells:
        cells = rg.permutation(cells)
    cells = np.array(cells)
    if permute_genes:
        genes = rg.permutation(genes)
    genes = np.array(genes)

    gene_cuts = [*np.arange(0, len(genes), step=n_genes_step)]
    if use_all:
        gene_cuts.append(len(genes))
    gene_bins = [genes[a:b] for a, b in zip(gene_cuts[:-1], gene_cuts[1:])]

    cell_cuts = [*np.arange(0, len(cells), step=n_cells_step)]
    if use_all:
        cell_cuts.append(len(cells))
    cell_bins = [cells[a:b] for a, b in zip(cell_cuts[:-1], cell_cuts[1:])]

    bins = []
    for cells_oi, genes_oi in itertools.product(cell_bins, gene_bins):
        bins.append(Minibatch(cells_oi=cells_oi, genes_oi=genes_oi, **kwargs))
    return bins


def create_bins_random(
    cells,
    genes,
    n_genes_total,
    n_genes_step=300,
    n_cells_step=1000,
    use_all=False,
    rg=None,
    permute_genes=True,
    **kwargs
):
    """
    Creates bins of cellxgene
    Within each cell bin, the genes are put in a new set of random bins
    """
    if rg is None:
        rg = np.random.RandomState()
    cells = rg.permutation(cells)

    cell_cuts = [*np.arange(0, len(cells), step=n_cells_step)]
    if use_all:
        cell_cuts.append(len(cells))
    cell_bins = [cells[a:b] for a, b in zip(cell_cuts[:-1], cell_cuts[1:])]

    bins = []
    for cells_oi in cell_bins:
        if permute_genes:
            genes = rg.permutation(genes)
        gene_cuts = [*np.arange(0, len(genes), step=n_genes_step)]
        if use_all:
            gene_cuts.append(len(genes))
        gene_bins = [genes[a:b] for a, b in zip(gene_cuts[:-1], gene_cuts[1:])]

        for genes_oi in gene_bins:
            bins.append(
                Minibatch(
                    cells_oi=cells_oi,
                    genes_oi=genes_oi,
                    **kwargs,
                )
            )
    return bins


class MinibatchCreator:
    def __init__(self, cells, genes, n_total_genes, n_cells_step, n_genes_step):
        self.cells = cells
        self.genes = genes
        self.n_total_genes = n_total_genes
        self.n_cells_step = n_cells_step
        self.n_genes_step = n_genes_step

    def create_minibatches(self, *args, **kwargs):
        minibatches = create_bins_ordered(
            self.cells,
            self.genes,
            self.n_total_genes,
            n_genes_step=self.n_genes_step,
            n_cells_step=self.n_cells_step,
            *args,
            **kwargs,
        )
        return minibatches


def get_minibatches_training(fold, n_genes, n_genes_step=500, n_cells_step=200):
    minibatcher = MinibatchCreator(
        fold["cells_train"],
        np.arange(n_genes),
        n_genes,
        n_genes_step=n_genes_step,
        n_cells_step=n_cells_step,
    )
    minibatches_train_sets = [
        {
            "tasks": minibatcher.create_minibatches(
                use_all=True, rg=np.random.RandomState(i)
            )
        }
        for i in range(10)
    ]
    fold["train"] = minibatches_train_sets

    rg = np.random.RandomState(0)
    fold["validation"] = create_bins_ordered(
        fold["cells_validation"],
        np.arange(n_genes),
        n_cells_step=n_cells_step,
        n_genes_step=n_genes_step,
        n_genes_total=n_genes,
        use_all=True,
        rg=rg,
    )
    fold["cellxgene_batch_size"] = n_cells_step * n_genes_step
    return fold


class Minibatcher:
    def __init__(
        self,
        cells,
        genes,
        n_cells_step,
        n_genes_step,
        use_all_cells=False,
        use_all_genes=True,
        permute_cells=True,
        permute_genes=False,
    ):
        self.cells = cells
        if not isinstance(genes, np.ndarray):
            genes = np.array(genes)
        self.genes = genes
        self.n_genes = len(genes)
        self.n_cells_step = n_cells_step
        self.n_genes_step = n_genes_step

        self.permute_cells = permute_cells
        self.permute_genes = permute_genes

        self.use_all_cells = use_all_cells or len(cells) < n_cells_step
        self.use_all_genes = use_all_genes or len(genes) < n_genes_step

        self.cellxgene_batch_size = n_cells_step * n_genes_step

        # calculate length
        n_cells = len(cells)
        n_genes = len(genes)
        if self.use_all_cells:
            n_cell_bins = math.ceil(n_cells / n_cells_step)
        else:
            n_cell_bins = math.floor(n_cells / n_cells_step)
        if self.use_all_genes:
            n_gene_bins = math.ceil(n_genes / n_genes_step)
        else:
            n_gene_bins = math.floor(n_genes / n_genes_step)
        self.length = n_cell_bins * n_gene_bins

        self.i = 0

    def __len__(self):
        return self.length

    def __iter__(self):
        self.rg = np.random.RandomState(self.i)

        if self.permute_cells:
            cells = self.rg.permutation(self.cells)
        else:
            cells = self.cells
        if self.permute_genes:
            genes = self.rg.permutation(self.genes)
        else:
            genes = self.genes

        gene_cuts = [*np.arange(0, len(genes), step=self.n_genes_step)]
        if self.use_all_genes:
            gene_cuts.append(len(genes))
        gene_bins = [genes[a:b] for a, b in zip(gene_cuts[:-1], gene_cuts[1:])]

        cell_cuts = [*np.arange(0, len(cells), step=self.n_cells_step)]
        if self.use_all_cells:
            cell_cuts.append(len(cells))
        cell_bins = [cells[a:b] for a, b in zip(cell_cuts[:-1], cell_cuts[1:])]

        for cells_oi, genes_oi in itertools.product(cell_bins, gene_bins):
            yield Minibatch(cells_oi=cells_oi, genes_oi=genes_oi)

        self.i += 1



import pathlib
import torch
import pickle
import numpy as np
import gzip
import copy
import json
import importlib
import shutil


class Flow:
    path: pathlib.Path
    default_name = None

    def __init__(self, path=None, folder=None, name=None, reset=False):
        if path is None:
            if folder is None:
                raise ValueError("Either path or folder must be specified")
            if name is None:
                if self.default_name is None:
                    raise ValueError(
                        "Cannot create Flow without name, and no default_name specified"
                    )
                name = self.default_name

            path = folder / name

        self.path = path
        if not path.exists():
            path.mkdir(parents=True, exist_ok=True)

        if reset:
            self.reset()

        if not self._get_info_path().exists():
            self._store_info()

    @classmethod
    def create(cls, path, **kwargs):
        path = pathlib.Path(path)
        self = cls(path=path)

        for key, value in kwargs.items():
            setattr(self, key, value)
        return self

    def _store_info(self):
        info = {"module": self.__class__.__module__, "class": self.__class__.__name__}

        json.dump(info, self._get_info_path().open("w"))

    def _get_info_path(self):
        return self.path / ".flow"

    def __repr__(self):
        return f'{self.__class__.__name__}("{self.path}")'

    @classmethod
    def from_path(cls, path):
        info = json.load(open(path / ".flow"))

        # load class
        module = importlib.import_module(info["module"])
        cls = getattr(module, info["class"])

        return cls(path=path)

    def reset(self):
        shutil.rmtree(self.path)
        self.path.mkdir(parents=True, exist_ok=True)


class Linked:
    def __init__(self, name):
        self.name = name

    def __get__(self, obj, type=None):
        if obj is not None:
            name = "_" + self.name
            if not hasattr(obj, name):
                path = obj.path / self.name

                if not path.exists():
                    raise FileNotFoundError(f"File {path} does not exist")
                if not path.is_symlink():
                    raise FileNotFoundError(f"File {path} is not a symlink")

                value = Flow.from_path(path.resolve())
                setattr(obj, name, value)

            return getattr(obj, name)

    def __set__(self, obj, value):
        # symlink to value.path
        path = obj.path / self.name
        name = "_" + self.name
        if path.exists():
            if not path.is_symlink():
                raise FileExistsError(f"File {path} already exists")
            else:
                path.unlink()
        path.symlink_to(value.path.resolve())
        setattr(obj, name, value)


class Stored:
    def __init__(self, name, default=None):
        self.name = name
        self.default = default

    def get_path(self, folder):
        return folder / (self.name + ".pkl")

    def __get__(self, obj, type=None):
        if obj is not None:
            name = "_" + self.name
            if not hasattr(obj, name):
                path = self.get_path(obj.path)
                if not path.exists():
                    if self.default is None:
                        raise FileNotFoundError(f"File {path} does not exist")
                    else:
                        value = self.default()
                        pickle.dump(value, path.open("wb"))
                setattr(obj, name, pickle.load(self.get_path(obj.path).open("rb")))
            return getattr(obj, name)

    def __set__(self, obj, value):
        name = "_" + self.name
        pickle.dump(value, self.get_path(obj.path).open("wb"))
        setattr(obj, name, value)


class StoredTorchInt64(Stored):
    def __get__(self, obj, type=None):
        if obj is not None:
            name = "_" + self.name
            if not hasattr(obj, name):
                x = pickle.load(self.get_path(obj.path).open("rb"))
                if not x.dtype is torch.int64:
                    x = x.to(torch.int64)
                if not x.is_contiguous():
                    x = x.contiguous()
                setattr(obj, name, x)
            return getattr(obj, name)

    def __set__(self, obj, value):
        value = value.to(torch.int64).contiguous()
        name = "_" + self.name
        pickle.dump(value, self.get_path(obj.path).open("wb"))
        setattr(obj, name, value)


class StoredTorchInt32(Stored):
    def __get__(self, obj, type=None):
        if obj is not None:
            name = "_" + self.name
            if not hasattr(obj, name):
                x = pickle.load(self.get_path(obj.path).open("rb"))
                if not x.dtype is torch.int32:
                    x = x.to(torch.int32)
                if not x.is_contiguous():
                    x = x.contiguous()
                setattr(obj, name, x)
            return getattr(obj, name)

    def __set__(self, obj, value):
        value = value.to(torch.int32).contiguous()
        name = "_" + self.name
        pickle.dump(value, self.get_path(obj.path).open("wb"))
        setattr(obj, name, value)


class StoredNumpyInt64(Stored):
    def __get__(self, obj, type=None):
        if obj is not None:
            name = "_" + self.name
            if not hasattr(obj, name):
                x = pickle.load(self.get_path(obj.path).open("rb"))
                if not x.dtype is np.int64:
                    x = x.astype(np.int64)
                if not x.flags["C_CONTIGUOUS"]:
                    x = np.ascontiguousarray(x)
                setattr(obj, name, x)
            return getattr(obj, name)

    def __set__(self, obj, value):
        value = np.ascontiguousarray(value.astype(np.int64))
        name = "_" + self.name
        pickle.dump(value, self.get_path(obj.path).open("wb"))
        setattr(obj, name, value)


class CompressedNumpy(Stored):
    dtype = np.float64

    def __get__(self, obj, type=None):
        if obj is not None:
            name = "_" + self.name
            if not hasattr(obj, name):
                x = pickle.load(
                    gzip.GzipFile(self.get_path(obj.path), "rb", compresslevel=3)
                )
                if not x.dtype is self.dtype:
                    x = x.astype(self.dtype)
                if not x.flags["C_CONTIGUOUS"]:
                    x = np.ascontiguousarray(x)
                setattr(obj, name, x)
            return getattr(obj, name)

    def __set__(self, obj, value):
        value = np.ascontiguousarray(value.astype(self.dtype))
        name = "_" + self.name
        pickle.dump(
            value, gzip.GzipFile(self.get_path(obj.path), "wb", compresslevel=3)
        )
        setattr(obj, name, value)


class CompressedNumpyFloat64(CompressedNumpy):
    dtype = np.float64


class CompressedNumpyInt64(CompressedNumpy):
    dtype = np.int64


class TSV(Stored):
    def __init__(self, name, columns=None):
        super().__init__(name)
        self.columns = columns

    def get_path(self, folder):
        return folder / (self.name + ".tsv")

    def __get__(self, obj=None, type=None):
        if obj is not None:
            name = "_" + self.name
            if not hasattr(obj, name):
                import pandas as pd

                x = pd.read_table(self.get_path(obj.path), index_col=0)
                setattr(obj, name, x)
            return getattr(obj, name)

    def __set__(self, obj, value, folder=None):
        name = "_" + self.name
        if folder is None:
            folder = obj.path
        value.to_csv(self.get_path(folder), sep="\t")
        setattr(obj, name, value)


class StoredDict:
    def __init__(self, name, cls):
        self.name = name
        self.cls = cls

    def get_path(self, folder):
        return folder / self.name

    def __get__(self, obj, type=None):
        return StoredDictInstance(self.name, self.get_path(obj.path), self.cls, obj)


class StoredDictInstance:
    def __init__(self, name, path, cls, obj):
        self.dict = {}
        self.cls = cls
        self.obj = obj
        self.name = name
        self.path = path
        if not self.path.exists():
            self.path.mkdir(parents=True)
        for file in self.path.iterdir():
            if file.is_dir():
                raise ValueError(f"Folder {file} in {self.obj.path} is not allowed")
            # key is file name without extension
            key = file.name.split(".")[0]
            self.dict[key] = self.cls(key)

    def __getitem__(self, key):
        return self.dict[key].__get__(self)

    def __setitem__(self, key, value):
        if key not in self.dict:
            self.dict[key] = self.cls(key)
        self.dict[key].__set__(self, value)



import tqdm.auto as tqdm
import torch
import pandas as pd
from chromatinhd.utils.ansi import colorcodes


class Trace:
    def __init__(self, n_significant_digits=3):
        self.train_steps = []
        self.validation_steps = []
        self.n_current_train_steps = 0
        self.n_current_validation_steps = 0
        self.n_last_train_steps = None
        self.n_last_validation_steps = None
        self.current_checkpoint = 0
        self.last_validation_diff = []

    def append(self, loss, epoch, step, phase="train"):
        if phase == "train":
            self.train_steps.append(
                {
                    "loss": loss,
                    "epoch": epoch,
                    "step": step,
                    "checkpoint": self.current_checkpoint,
                }
            )
            self.n_current_train_steps += 1
        else:
            self.validation_steps.append(
                {
                    "loss": loss,
                    "epoch": epoch,
                    "step": step,
                    "checkpoint": self.current_checkpoint,
                }
            )
            self.n_current_validation_steps += 1

    def checkpoint(self, logger=print):
        if (
            (self.n_last_train_steps is not None)
            and (self.n_last_train_steps > 0)
            and (self.n_current_train_steps > 0)
        ):
            last_train_steps = pd.DataFrame(
                self.train_steps[
                    -(self.n_current_train_steps + self.n_last_train_steps) : -(
                        self.n_current_train_steps
                    )
                ]
            )
            current_train_steps = pd.DataFrame(
                self.train_steps[-(self.n_current_train_steps) :]
            )

            current_loss = current_train_steps["loss"].mean()
            diff_loss = (
                current_train_steps["loss"].mean() - last_train_steps["loss"].mean()
            )
            perc_diff_loss = diff_loss / current_loss

            logger.info(
                f"{'train':>10} {current_loss:+.2f} Δ{colorcodes.color_sign(diff_loss, '{:+.3f}')} {perc_diff_loss:+.2%}"
            )
        self.n_last_train_steps = self.n_current_train_steps
        self.n_current_train_steps = 0

        if len(self.validation_steps) > 0:
            current_validation_steps = pd.DataFrame(
                self.validation_steps[-(self.n_current_validation_steps) :]
            )
            current_loss = current_validation_steps["loss"].mean()
            if (self.n_last_validation_steps is not None) and (
                self.n_last_validation_steps > 0
            ):
                if self.n_current_validation_steps == 0:
                    raise ValueError(
                        "No validation steps were run since last checkpoint"
                    )
                assert len(self.validation_steps) >= (
                    self.n_current_validation_steps + self.n_last_validation_steps
                )

                last_validation_steps = pd.DataFrame(
                    self.validation_steps[
                        -(
                            self.n_current_validation_steps
                            + self.n_last_validation_steps
                        ) : -(self.n_current_validation_steps)
                    ]
                )

                diff_loss = (
                    current_validation_steps["loss"].mean()
                    - last_validation_steps["loss"].mean()
                )
                self.last_validation_diff.append(diff_loss)
                perc_diff_loss = diff_loss / current_loss

                logger.info(
                    f"{'validation':>10} {current_loss:+.2f} Δ{colorcodes.color_sign(diff_loss, '{:+.3f}')} {perc_diff_loss:+.2%}"
                )
            else:
                logger.info(f"{'validation':>10} {current_loss:+.2f}")
            self.n_last_validation_steps = self.n_current_validation_steps
            self.n_current_validation_steps = 0

        self.current_checkpoint += 1

    def plot(self):
        import matplotlib.pyplot as plt

        fig, ax = plt.subplots()
        plotdata_validation = (
            pd.DataFrame(self.validation_steps)
            .groupby("checkpoint")
            .mean()
            .reset_index()
        )
        plotdata_train = (
            pd.DataFrame(self.train_steps).groupby("checkpoint").mean().reset_index()
        )
        ax.plot(
            plotdata_validation["checkpoint"],
            plotdata_validation["loss"],
            label="validation",
        )
        ax.plot(plotdata_train["checkpoint"], plotdata_train["loss"], label="train")
        ax.legend()



import pandas as pd
import numpy as np
import torch


def select_cutwindow(coordinates, window_start, window_end):
    """
    check whether coordinate 0 or coordinate 1 is within the window
    """
    return ~(
        ((coordinates[:, 0] < window_end) & (coordinates[:, 0] > window_start))
        | ((coordinates[:, 1] < window_end) & (coordinates[:, 1] > window_start))
    )


class WindowCensorer:
    def __init__(self, window, window_size=100):
        design = [{"window": "control"}]
        cuts = np.arange(*window, step=int(window_size))

        for window_start, window_end in zip(cuts, cuts + window_size):
            if window_start < window[0]:
                continue
            if window_end > window[1]:
                continue
            design.append(
                {
                    "window_start": window_start,
                    "window_end": window_end,
                    "window_mid": window_start + (window_end - window_start) / 2,
                    "window": f"{window_start}-{window_end}",
                }
            )
        design = pd.DataFrame(design).set_index("window")
        assert design.index.is_unique
        design["ix"] = np.arange(len(design))
        self.design = design

    def __len__(self):
        return len(self.design)

    def __call__(self, data):
        for window_start, window_end in zip(
            self.design["window_start"], self.design["window_end"]
        ):
            if np.isnan(window_start):
                fragments_oi = None
            else:
                fragments_oi = select_cutwindow(
                    data.fragments.coordinates, window_start, window_end
                )
            yield fragments_oi


class MultiWindowCensorer:
    def __init__(self, window, window_sizes=(50, 100, 200, 500), relative_stride=0.5):
        design = [{"window": "control"}]
        for window_size in window_sizes:
            cuts = np.arange(*window, step=int(window_size * relative_stride))

            for window_start, window_end in zip(cuts, cuts + window_size):
                if window_start < window[0]:
                    continue
                if window_end > window[1]:
                    continue
                design.append(
                    {
                        "window_start": window_start,
                        "window_end": window_end,
                        "window_mid": window_start + (window_end - window_start) / 2,
                        "window_size": window_size,
                        "window": f"{window_start}-{window_end}",
                    }
                )
        design = pd.DataFrame(design).set_index("window")
        assert design.index.is_unique
        design["ix"] = np.arange(len(design))
        self.design = design

    def __len__(self):
        return len(self.design)

    def __call__(self, data):
        for window_start, window_end in zip(
            self.design["window_start"], self.design["window_end"]
        ):
            if np.isnan(window_start):
                fragments_oi = None
            else:
                fragments_oi = select_cutwindow(
                    data.fragments.coordinates, window_start, window_end
                )
            yield fragments_oi



import chromatinhd as chd
import numpy as np
import pandas as pd
import xarray as xr
import pickle
import scipy.stats
import tqdm.auto as tqdm


def fdr(p_vals):
    from scipy.stats import rankdata

    ranked_p_values = rankdata(p_vals)
    fdr = p_vals * len(p_vals) / ranked_p_values
    fdr[fdr > 1] = 1

    return fdr


class GeneMultiWindow(chd.flow.Flow):
    design = chd.flow.Stored("design")

    genes = chd.flow.Stored("genes", default=set)

    def score(
        self,
        fragments,
        transcriptome,
        models,
        folds,
        genes,
        censorer,
        force=False,
        device="cuda",
    ):
        force_ = force
        design = censorer.design.iloc[1:].copy()
        self.design = design

        pbar = tqdm.tqdm(genes, leave=False)

        for gene in pbar:
            pbar.set_description(gene)
            scores_file = self.get_scoring_path(gene) / "scores.pkl"

            force = force_
            if not scores_file.exists():
                force = True

            if force:
                deltacor_folds = []
                lost_folds = []
                effect_folds = []
                for fold, model in zip(folds, models):
                    predicted, expected, n_fragments = model.get_prediction_censored(
                        fragments,
                        transcriptome,
                        censorer,
                        cell_ixs=np.concatenate(
                            [fold["cells_validation"], fold["cells_test"]]
                        ),
                        genes=[gene],
                        device=device,
                    )

                    # select 1st gene, given that we're working with one gene anyway
                    predicted = predicted[..., 0]
                    expected = expected[..., 0]
                    n_fragments = n_fragments[..., 0]

                    cor = chd.utils.paircor(predicted, expected, dim=-1)
                    deltacor = cor[1:] - cor[0]

                    lost = (n_fragments[0] - n_fragments[1:]).mean(-1)

                    effect = (predicted[0] - predicted[1:]).mean(-1)

                    deltacor_folds.append(deltacor)
                    lost_folds.append(lost)
                    effect_folds.append(effect)

                deltacor_folds = np.stack(deltacor_folds, 0)
                lost_folds = np.stack(lost_folds, 0)
                effect_folds = np.stack(effect_folds, 0)

                result = xr.Dataset(
                    {
                        "deltacor": xr.DataArray(
                            deltacor_folds,
                            coords=[
                                ("model", np.arange(len(models))),
                                ("window", design.index),
                            ],
                        ),
                        "lost": xr.DataArray(
                            lost_folds,
                            coords=[
                                ("model", np.arange(len(models))),
                                ("window", design.index),
                            ],
                        ),
                        "effect": xr.DataArray(
                            effect_folds,
                            coords=[
                                ("model", np.arange(len(models))),
                                ("window", design.index),
                            ],
                        ),
                    }
                )

                pickle.dump(result, scores_file.open("wb"))

                self.genes = self.genes | {gene}

    def interpolate(self, genes=None, force=False):
        force_ = force

        if genes is None:
            genes = self.genes

        pbar = tqdm.tqdm(genes, leave=False)

        for gene in pbar:
            pbar.set_description(gene)
            scores_file = self.get_scoring_path(gene) / "scores.pkl"

            if not scores_file.exists():
                continue

            interpolate_file = self.get_scoring_path(gene) / "interpolated.pkl"

            force = force_
            if not interpolate_file.exists():
                force = True

            if force:
                scores = pickle.load(scores_file.open("rb"))
                x = scores["deltacor"].values
                scores_statistical = []
                for i in range(x.shape[1]):
                    scores_statistical.append(
                        scipy.stats.ttest_1samp(x[:, i], 0, alternative="less").pvalue
                    )
                scores_statistical = pd.DataFrame({"pvalue": scores_statistical})
                scores_statistical["qval"] = fdr(scores_statistical["pvalue"])

                plotdata = scores.mean("model").stack().to_dataframe()
                plotdata = self.design.join(plotdata)

                plotdata["qval"] = scores_statistical["qval"].values

                window_sizes_info = pd.DataFrame(
                    {"window_size": self.design["window_size"].unique()}
                ).set_index("window_size")
                window_sizes_info["ix"] = np.arange(len(window_sizes_info))

                # interpolate
                positions_oi = np.arange(
                    self.design["window_start"].min(),
                    self.design["window_end"].max() + 1,
                )

                deltacor_interpolated = np.zeros(
                    (len(window_sizes_info), len(positions_oi))
                )
                lost_interpolated = np.zeros(
                    (len(window_sizes_info), len(positions_oi))
                )
                effect_interpolated = np.zeros(
                    (len(window_sizes_info), len(positions_oi))
                )
                for window_size, window_size_info in window_sizes_info.iterrows():
                    plotdata_oi = plotdata.query("window_size == @window_size")
                    x = plotdata_oi["window_mid"].values.copy()
                    y = plotdata_oi["deltacor"].values.copy()
                    y[plotdata_oi["qval"] > 0.1] = 0.0
                    deltacor_interpolated_ = np.clip(
                        np.interp(positions_oi, x, y) / window_size * 1000,
                        -np.inf,
                        0,
                        # np.inf,
                    )
                    deltacor_interpolated[
                        window_size_info["ix"], :
                    ] = deltacor_interpolated_

                    lost_interpolated_ = (
                        np.interp(
                            positions_oi, plotdata_oi["window_mid"], plotdata_oi["lost"]
                        )
                        / window_size
                        * 1000
                    )
                    lost_interpolated[window_size_info["ix"], :] = lost_interpolated_

                    effect_interpolated_ = (
                        np.interp(
                            positions_oi,
                            plotdata_oi["window_mid"],
                            plotdata_oi["effect"],
                        )
                        / window_size
                        * 1000
                    )
                    effect_interpolated[
                        window_size_info["ix"], :
                    ] = effect_interpolated_

                deltacor = xr.DataArray(
                    deltacor_interpolated.mean(0),
                    coords=[
                        ("position", positions_oi),
                    ],
                )
                lost = xr.DataArray(
                    lost_interpolated.mean(0),
                    coords=[
                        ("position", positions_oi),
                    ],
                )

                effect = xr.DataArray(
                    effect_interpolated.mean(0),
                    coords=[
                        ("position", positions_oi),
                    ],
                )

                # save
                interpolated = xr.Dataset(
                    {"deltacor": deltacor, "lost": lost, "effect": effect}
                )
                pickle.dump(
                    interpolated,
                    interpolate_file.open("wb"),
                )

    def get_plotdata(self, gene):
        interpolated_file = self.get_scoring_path(gene) / "interpolated.pkl"
        if not interpolated_file.exists():
            raise FileNotFoundError(f"File {interpolated_file} does not exist")

        interpolated = pickle.load(interpolated_file.open("rb"))

        plotdata = interpolated.to_dataframe()

        return plotdata

    def get_scoring_path(self, gene):
        path = self.path / f"{gene}"
        path.mkdir(parents=True, exist_ok=True)
        return path



import numpy as np
import chromatinhd as chd


class TestTranscriptomeFragments:
    def test_example(self, example_fragments, example_transcriptome):
        loader = chd.models.pred.loader.TranscriptomeFragments(
            fragments=example_fragments,
            transcriptome=example_transcriptome,
            cellxgene_batch_size=10000,
        )

        minibatch = chd.models.pred.loader.Minibatch(
            cells_oi=np.arange(20), genes_oi=np.arange(5), phase="train"
        )
        result = loader.load(minibatch)



import numpy as np
import pandas as pd
import pickle

from chromatinhd.flow import (
    Flow,
    StoredTorchInt32,
    Stored,
    StoredTorchInt64,
    TSV,
    Linked,
)

from chromatinhd.data.fragments import Fragments

import torch
import math
import pathlib
import typing
import tqdm


class Folds(Flow):
    """
    Folds of multiple cell and gene combinations
    """

    folds = Stored("folds")
    """the folds"""

    def sample_cells(self, fragments: Fragments, n_folds: int, n_repeats: int = 1):
        folds = []

        for repeat_ix in range(n_repeats):
            generator = np.random.RandomState(repeat_ix)

            cells_all = generator.permutation(fragments.n_cells)

            cell_bins = np.floor(
                (np.arange(len(cells_all)) / (len(cells_all) / n_folds))
            )

            for i in range(n_folds):
                cells_train = cells_all[cell_bins != i]
                cells_validation_test = cells_all[cell_bins == i]
                cells_validation = cells_validation_test[
                    : (len(cells_validation_test) // 2)
                ]
                cells_test = cells_validation_test[(len(cells_validation_test) // 2) :]

                folds.append(
                    {
                        "cells_train": cells_train,
                        "cells_validation": cells_validation,
                        "cells_test": cells_test,
                        "repeat": repeat_ix,
                    }
                )
        self.folds = folds

    def __getitem__(self, ix):
        return self.folds[ix]

    def __len__(self):
        return len(self.folds)



import numpy as np
import pandas as pd

import scanpy as sc

import collections
import subprocess as sp
import tqdm.auto as tqdm
import pathlib
import pickle

from chromatinhd.flow import Flow

htslib_folder = pathlib.Path("/data/peak_free_atac/software/htslib-1.16/")
tabix_location = htslib_folder / "tabix"


class PeakCounts(Flow):
    def create_adata(self, original_adata):
        adata = sc.AnnData(self.counts, obs=self.obs, var=self.var)
        adata.obsm["X_umap"] = original_adata.obsm["X_umap"]
        self.adata = adata

    def count_peaks(self, fragments_location, cell_ids):
        # extract unique peaks
        peaks = self.peaks
        unique_peak_ids = list(set(peaks["peak"]))
        peak_idxs = {peak_id: i for i, peak_id in enumerate(unique_peak_ids)}

        # create peaks file for tabix
        self.peaks_bed = peaks[["chrom", "start", "end"]]

        # count
        counts = collections.defaultdict(int)
        cell_ids = [str(cell_id) for cell_id in cell_ids]
        barcode_idxs = {barcode: ix for ix, barcode in enumerate(cell_ids)}

        process = sp.Popen(
            [
                tabix_location,
                fragments_location,
                "-R",
                self.peaks_bed_path,
                "--separate-regions",
            ],
            stdout=sp.PIPE,
        )
        counter = tqdm.tqdm(total=len(unique_peak_ids), smoothing=0)
        for line in process.stdout:
            line = line.decode("utf-8")
            if line.startswith("#"):
                peak = line.rstrip("\n").lstrip("#")
                peak_idx = peak_idxs[peak]
                counter.update(1)
            else:
                fragment = line.split("\t")
                barcode = fragment[3].strip("\n")

                if barcode in barcode_idxs:
                    counts[(barcode_idxs[barcode], peak_idx)] += 1

        # convert to sparse
        import scipy.sparse

        i = [k[0] for k in counts.keys()]
        j = [k[1] for k in counts.keys()]
        v = [v for v in counts.values()]
        counts_csr = scipy.sparse.csr_matrix(
            (v, (i, j)), shape=(len(barcode_idxs), len(peak_idxs))
        )

        if counts_csr.sum() == 0:
            raise ValueError("Something went wrong with counting")

        self.counts = counts_csr

        # create obs
        obs = pd.DataFrame(
            {"cell": list(barcode_idxs.keys()), "ix": list(barcode_idxs.values())}
        ).set_index("cell")
        self.obs = obs.copy()

        # create var
        var = pd.DataFrame(index=pd.Series(unique_peak_ids, name="peak"))
        var["ix"] = np.arange(var.shape[0])
        self.var = var

    @property
    def peaks_bed_path(self):
        return self.path / "peaks_bed.tsv"

    @property
    def peaks_bed(self):
        return pd.read_table(self.peaks_bed_path)

    @peaks_bed.setter
    def peaks_bed(self, value):
        value.to_csv(self.peaks_bed_path, sep="\t", header=False, index=False)

    @property
    def peaks(self):
        return pd.read_table(self.path / "peaks.tsv", index_col=0)

    @peaks.setter
    def peaks(self, value):
        value.index.name = "peak_gene"
        value.to_csv(self.path / "peaks.tsv", sep="\t")

    @property
    def counts(self):
        return pickle.load((self.path / "counts.pkl").open("rb"))

    @counts.setter
    def counts(self, value):
        pickle.dump(value, (self.path / "counts.pkl").open("wb"))

    @property
    def var(self):
        return pd.read_table(self.path / "var.tsv", index_col=0)

    @var.setter
    def var(self, value):
        value.index.name = "peak"
        value.to_csv(self.path / "var.tsv", sep="\t")

    @property
    def obs(self):
        return pd.read_table(self.path / "obs.tsv", index_col=0)

    @obs.setter
    def obs(self, value):
        value.index.name = "peak"
        value.to_csv(self.path / "obs.tsv", sep="\t")

    _adata = None

    @property
    def adata(self):
        if self._adata is None:
            self._adata = pickle.load((self.path / "adata.pkl").open("rb"))
        return self._adata

    @adata.setter
    def adata(self, value):
        pickle.dump(value, (self.path / "adata.pkl").open("wb"))
        self._adata = value


class FullPeak(PeakCounts):
    default_name = "full_peak"

    def create_peaks(self, original_peak_annot):
        original_peak_annot.index = pd.Index(
            original_peak_annot.chrom
            + ":"
            + original_peak_annot.start.astype(str)
            + "-"
            + original_peak_annot.end.astype(str)
        )

        peaks = original_peak_annot.copy()

        peaks.index = pd.Index(
            peaks.chrom + ":" + peaks.start.astype(str) + "-" + peaks.end.astype(str),
            name="peak",
        )
        peaks = peaks.groupby(level=0).first()
        peaks.index = pd.Index(
            peaks.chrom + ":" + peaks.start.astype(str) + "-" + peaks.end.astype(str),
            name="peak",
        )

        peaks["original_peak"] = peaks.index

        self.peaks = peaks


class HalfPeak(PeakCounts):
    default_name = "half_peak"

    def create_peaks(self, original_peak_annot):
        original_peak_annot.index = pd.Index(
            original_peak_annot.chrom
            + ":"
            + original_peak_annot.start.astype(str)
            + "-"
            + original_peak_annot.end.astype(str)
        )

        peaks = []
        for peak_id, peak in tqdm.tqdm(
            original_peak_annot.iterrows(), total=original_peak_annot.shape[0]
        ):
            peak1 = peak.copy()
            peak1["end"] = int(peak["start"] + int(peak["end"] - peak["start"]) / 2)
            peak1["original_peak"] = peak_id
            peaks.append(peak1)
            peak2 = peak.copy()
            peak2["start"] = int(peak["end"] - int(peak["end"] - peak["start"]) / 2)
            peak2["original_peak"] = peak_id
            peaks.append(peak2)

        peaks = pd.DataFrame(peaks)
        peaks.index = pd.Index(
            peaks.chrom + ":" + peaks.start.astype(str) + "-" + peaks.end.astype(str)
        )
        peaks = peaks.groupby(level=0).first()
        peaks.index = pd.Index(
            peaks.chrom + ":" + peaks.start.astype(str) + "-" + peaks.end.astype(str)
        )

        self.peaks = peaks


class ThirdPeak(PeakCounts):
    default_name = "third_peak"

    def create_peaks(self, original_peak_annot):
        original_peak_annot.index = pd.Index(
            original_peak_annot.chrom
            + ":"
            + original_peak_annot.start.astype(str)
            + "-"
            + original_peak_annot.end.astype(str)
        )

        peaks = []
        for peak_id, peak in original_peak_annot.iterrows():
            scale = peak["end"] - peak["start"]
            peak1 = peak.copy()
            peak1["end"] = int(peak["start"] + scale / 3)
            peak1["original_peak"] = peak_id
            peaks.append(peak1)
            peak2 = peak.copy()
            peak2["start"] = int(peak["start"] + scale / 3)
            peak2["end"] = int(peak["end"] - scale / 3)
            peak2["original_peak"] = peak_id
            peaks.append(peak2)
            peak2 = peak.copy()
            peak2["start"] = int(peak["end"] - scale / 3)
            peak2["original_peak"] = peak_id
            peaks.append(peak2)

        peaks = pd.DataFrame(peaks)
        peaks.index = pd.Index(
            peaks.chrom + ":" + peaks.start.astype(str) + "-" + peaks.end.astype(str)
        )
        peaks = peaks.groupby(level=0).first()
        peaks.index = pd.Index(
            peaks.chrom + ":" + peaks.start.astype(str) + "-" + peaks.end.astype(str)
        )

        self.peaks = peaks


class BroaderPeak(PeakCounts):
    default_name = "broader_peak"

    def create_peaks(self, original_peak_annot):
        original_peak_annot.index = pd.Index(
            original_peak_annot.chrom
            + ":"
            + original_peak_annot.start.astype(str)
            + "-"
            + original_peak_annot.end.astype(str)
        )

        peaks = original_peak_annot.copy()
        peaks["original_peak"] = peaks.index.copy()

        peaks["start"] = np.maximum(1, peaks["start"] - 5000)
        peaks["end"] = peaks["end"] + 5000

        peaks.index = pd.Index(
            peaks.chrom + ":" + peaks.start.astype(str) + "-" + peaks.end.astype(str)
        )
        peaks = peaks.groupby(level=0).first()
        peaks.index = pd.Index(
            peaks.chrom + ":" + peaks.start.astype(str) + "-" + peaks.end.astype(str)
        )

        self.peaks = peaks


class FragmentPeak(FullPeak):
    default_name = "fragment_peak"

    def count_peaks(self, fragments_location, cell_ids):
        # add ix to peaks
        peaks = self.peaks
        peaks["ix"] = np.arange(peaks.shape[0])

        # create peaks file for tabix
        peaks_bed = peaks[["chrom", "start", "end"]]
        self.peaks_bed = peaks_bed

        # count
        fragments = []

        peak_idxs = peaks["ix"].to_dict()
        barcode_idxs = {barcode: ix for ix, barcode in enumerate(cell_ids)}

        fragment_cutoffs = (125, 250, 400)

        process = sp.Popen(
            [
                tabix_location,
                fragments_location,
                "-R",
                peaks_bed_path,
                "--separate-regions",
            ],
            stdout=sp.PIPE,
        )
        counter = tqdm.tqdm(total=self.peaks.shape[0], smoothing=0)
        for line in process.stdout:
            line = line.decode("utf-8")
            if line.startswith("#"):
                peak = line.rstrip("\n").lstrip("#")
                peak_idx = peak_idxs[peak]
                counter.update(1)
            else:
                fragment = line.split("\t")
                barcode = fragment[3]
                if barcode in barcode_idxs:
                    length = int(fragment[2]) - int(fragment[1])
                    for i, cutoff in enumerate(fragment_cutoffs):
                        if length < cutoff:
                            fragments.append((barcode_idxs[barcode], peak_idx, i))
                            break
                    else:
                        fragments.append((barcode_idxs[barcode], peak_idx, i))

            # if counter.n > 1000:
            #     break

        # count individual fragments
        counts = collections.defaultdict(int)
        for k in fragments:
            counts[(k[0], k[1] * (len(fragment_cutoffs) + 1) + k[2])] += 1

        # convert to sparse
        import scipy.sparse

        i = [k[0] for k in counts.keys()]
        j = [k[1] for k in counts.keys()]
        v = [v for v in counts.values()]
        counts_csr = scipy.sparse.csr_matrix(
            (v, (i, j)),
            shape=(len(barcode_idxs), len(peak_idxs) * (len(fragment_cutoffs) + 1)),
        )

        self.store("counts", counts_csr)

        # create obs
        obs = pd.DataFrame(
            {"cell": list(barcode_idxs.keys()), "ix": list(barcode_idxs.values())}
        ).set_index("cell")
        self.store("obs", obs)

        # create var
        var = peaks.loc[peaks.index.repeat(len(fragment_cutoffs) + 1)].copy()
        var["fragment_bin"] = (list(fragment_cutoffs) + ["inf"]) * peaks.shape[0]
        var.index = var.index + "_" + var["fragment_bin"].astype(str)

        self.var = var



import math

import torch
from torch.nn import functional as F

import utils
from nde import transforms

DEFAULT_MIN_BIN_WIDTH = 1e-3
DEFAULT_MIN_BIN_HEIGHT = 1e-3
DEFAULT_EPS = 1e-5
DEFAULT_QUADRATIC_THRESHOLD = 1e-3


def unconstrained_cubic_spline(
    inputs,
    unnormalized_widths,
    unnormalized_heights,
    unnorm_derivatives_left,
    unnorm_derivatives_right,
    inverse=False,
    tail_bound=1.0,
    tails="linear",
    min_bin_width=DEFAULT_MIN_BIN_WIDTH,
    min_bin_height=DEFAULT_MIN_BIN_HEIGHT,
    eps=DEFAULT_EPS,
    quadratic_threshold=DEFAULT_QUADRATIC_THRESHOLD,
):
    inside_interval_mask = (inputs >= -tail_bound) & (inputs <= tail_bound)
    outside_interval_mask = ~inside_interval_mask

    outputs = torch.zeros_like(inputs)
    logabsdet = torch.zeros_like(inputs)

    if tails == "linear":
        outputs[outside_interval_mask] = inputs[outside_interval_mask]
        logabsdet[outside_interval_mask] = 0
    else:
        raise RuntimeError("{} tails are not implemented.".format(tails))

    outputs[inside_interval_mask], logabsdet[inside_interval_mask] = cubic_spline(
        inputs=inputs[inside_interval_mask],
        unnormalized_widths=unnormalized_widths[inside_interval_mask, :],
        unnormalized_heights=unnormalized_heights[inside_interval_mask, :],
        unnorm_derivatives_left=unnorm_derivatives_left[inside_interval_mask, :],
        unnorm_derivatives_right=unnorm_derivatives_right[inside_interval_mask, :],
        inverse=inverse,
        left=-tail_bound,
        right=tail_bound,
        bottom=-tail_bound,
        top=tail_bound,
        min_bin_width=min_bin_width,
        min_bin_height=min_bin_height,
        eps=eps,
        quadratic_threshold=quadratic_threshold,
    )

    return outputs, logabsdet


def cubic_spline(
    inputs,
    unnormalized_widths,
    unnormalized_heights,
    unnorm_derivatives_left,
    unnorm_derivatives_right,
    inverse=False,
    left=0.0,
    right=1.0,
    bottom=0.0,
    top=1.0,
    min_bin_width=DEFAULT_MIN_BIN_WIDTH,
    min_bin_height=DEFAULT_MIN_BIN_HEIGHT,
    eps=DEFAULT_EPS,
    quadratic_threshold=DEFAULT_QUADRATIC_THRESHOLD,
):
    """
    References:
    > Blinn, J. F. (2007). How to solve a cubic equation, part 5: Back to numerics. IEEE Computer
    Graphics and Applications, 27(3):78–89.
    """
    if not inverse and (torch.min(inputs) < left or torch.max(inputs) > right):
        raise transforms.InputOutsideDomain()
    elif inverse and (torch.min(inputs) < bottom or torch.max(inputs) > top):
        raise transforms.InputOutsideDomain()

    num_bins = unnormalized_widths.shape[-1]

    if min_bin_width * num_bins > 1.0:
        raise ValueError("Minimal bin width too large for the number of bins")
    if min_bin_height * num_bins > 1.0:
        raise ValueError("Minimal bin height too large for the number of bins")

    if inverse:
        inputs = (inputs - bottom) / (top - bottom)
    else:
        inputs = (inputs - left) / (right - left)

    widths = F.softmax(unnormalized_widths, dim=-1)
    widths = min_bin_width + (1 - min_bin_width * num_bins) * widths

    cumwidths = torch.cumsum(widths, dim=-1)
    cumwidths[..., -1] = 1
    cumwidths = F.pad(cumwidths, pad=(1, 0), mode="constant", value=0.0)

    heights = F.softmax(unnormalized_heights, dim=-1)
    heights = min_bin_height + (1 - min_bin_height * num_bins) * heights

    cumheights = torch.cumsum(heights, dim=-1)
    cumheights[..., -1] = 1
    cumheights = F.pad(cumheights, pad=(1, 0), mode="constant", value=0.0)

    slopes = heights / widths
    min_something_1 = torch.min(torch.abs(slopes[..., :-1]), torch.abs(slopes[..., 1:]))
    min_something_2 = (
        0.5
        * (widths[..., 1:] * slopes[..., :-1] + widths[..., :-1] * slopes[..., 1:])
        / (widths[..., :-1] + widths[..., 1:])
    )
    min_something = torch.min(min_something_1, min_something_2)

    derivatives_left = (
        torch.sigmoid(unnorm_derivatives_left) * 3 * slopes[..., 0][..., None]
    )
    derivatives_right = (
        torch.sigmoid(unnorm_derivatives_right) * 3 * slopes[..., -1][..., None]
    )

    derivatives = min_something * (
        torch.sign(slopes[..., :-1]) + torch.sign(slopes[..., 1:])
    )
    derivatives = torch.cat([derivatives_left, derivatives, derivatives_right], dim=-1)

    a = (derivatives[..., :-1] + derivatives[..., 1:] - 2 * slopes) / widths.pow(2)
    b = (3 * slopes - 2 * derivatives[..., :-1] - derivatives[..., 1:]) / widths
    c = derivatives[..., :-1]
    d = cumheights[..., :-1]

    if inverse:
        bin_idx = utils.searchsorted(cumheights, inputs)[..., None]
    else:
        bin_idx = utils.searchsorted(cumwidths, inputs)[..., None]

    inputs_a = a.gather(-1, bin_idx)[..., 0]
    inputs_b = b.gather(-1, bin_idx)[..., 0]
    inputs_c = c.gather(-1, bin_idx)[..., 0]
    inputs_d = d.gather(-1, bin_idx)[..., 0]

    input_left_cumwidths = cumwidths.gather(-1, bin_idx)[..., 0]
    input_right_cumwidths = cumwidths.gather(-1, bin_idx + 1)[..., 0]

    if inverse:
        # Modified coefficients for solving the cubic.
        inputs_b_ = (inputs_b / inputs_a) / 3.0
        inputs_c_ = (inputs_c / inputs_a) / 3.0
        inputs_d_ = (inputs_d - inputs) / inputs_a

        delta_1 = -inputs_b_.pow(2) + inputs_c_
        delta_2 = -inputs_c_ * inputs_b_ + inputs_d_
        delta_3 = inputs_b_ * inputs_d_ - inputs_c_.pow(2)

        discriminant = 4.0 * delta_1 * delta_3 - delta_2.pow(2)

        depressed_1 = -2.0 * inputs_b_ * delta_1 + delta_2
        depressed_2 = delta_1

        three_roots_mask = (
            discriminant >= 0
        )  # Discriminant == 0 might be a problem in practice.
        one_root_mask = discriminant < 0

        outputs = torch.zeros_like(inputs)

        # Deal with one root cases.

        p = utils.cbrt(
            (-depressed_1[one_root_mask] + torch.sqrt(-discriminant[one_root_mask]))
            / 2.0
        )
        q = utils.cbrt(
            (-depressed_1[one_root_mask] - torch.sqrt(-discriminant[one_root_mask]))
            / 2.0
        )

        outputs[one_root_mask] = (
            (p + q) - inputs_b_[one_root_mask] + input_left_cumwidths[one_root_mask]
        )

        # Deal with three root cases.

        theta = torch.atan2(
            torch.sqrt(discriminant[three_roots_mask]), -depressed_1[three_roots_mask]
        )
        theta /= 3.0

        cubic_root_1 = torch.cos(theta)
        cubic_root_2 = torch.sin(theta)

        root_1 = cubic_root_1
        root_2 = -0.5 * cubic_root_1 - 0.5 * math.sqrt(3) * cubic_root_2
        root_3 = -0.5 * cubic_root_1 + 0.5 * math.sqrt(3) * cubic_root_2

        root_scale = 2 * torch.sqrt(-depressed_2[three_roots_mask])
        root_shift = (
            -inputs_b_[three_roots_mask] + input_left_cumwidths[three_roots_mask]
        )

        root_1 = root_1 * root_scale + root_shift
        root_2 = root_2 * root_scale + root_shift
        root_3 = root_3 * root_scale + root_shift

        root1_mask = ((input_left_cumwidths[three_roots_mask] - eps) < root_1).float()
        root1_mask *= (root_1 < (input_right_cumwidths[three_roots_mask] + eps)).float()

        root2_mask = ((input_left_cumwidths[three_roots_mask] - eps) < root_2).float()
        root2_mask *= (root_2 < (input_right_cumwidths[three_roots_mask] + eps)).float()

        root3_mask = ((input_left_cumwidths[three_roots_mask] - eps) < root_3).float()
        root3_mask *= (root_3 < (input_right_cumwidths[three_roots_mask] + eps)).float()

        roots = torch.stack([root_1, root_2, root_3], dim=-1)
        masks = torch.stack([root1_mask, root2_mask, root3_mask], dim=-1)
        mask_index = torch.argsort(masks, dim=-1, descending=True)[..., 0][..., None]
        outputs[three_roots_mask] = torch.gather(roots, dim=-1, index=mask_index).view(
            -1
        )

        # Deal with a -> 0 (almost quadratic) cases.

        quadratic_mask = inputs_a.abs() < quadratic_threshold
        a = inputs_b[quadratic_mask]
        b = inputs_c[quadratic_mask]
        c = inputs_d[quadratic_mask] - inputs[quadratic_mask]
        alpha = (-b + torch.sqrt(b.pow(2) - 4 * a * c)) / (2 * a)
        outputs[quadratic_mask] = alpha + input_left_cumwidths[quadratic_mask]

        shifted_outputs = outputs - input_left_cumwidths
        logabsdet = -torch.log(
            (
                3 * inputs_a * shifted_outputs.pow(2)
                + 2 * inputs_b * shifted_outputs
                + inputs_c
            )
        )
    else:
        shifted_inputs = inputs - input_left_cumwidths
        outputs = (
            inputs_a * shifted_inputs.pow(3)
            + inputs_b * shifted_inputs.pow(2)
            + inputs_c * shifted_inputs
            + inputs_d
        )

        logabsdet = torch.log(
            (
                3 * inputs_a * shifted_inputs.pow(2)
                + 2 * inputs_b * shifted_inputs
                + inputs_c
            )
        )

    if inverse:
        outputs = outputs * (right - left) + left
        logabsdet = logabsdet - math.log(top - bottom) + math.log(right - left)
    else:
        outputs = outputs * (top - bottom) + bottom
        logabsdet = logabsdet + math.log(top - bottom) - math.log(right - left)

    return (outputs,)



import numpy as np
import pandas as pd
import pickle

from chromatinhd.flow import Flow, Stored, StoredDict
from chromatinhd import sparse
from chromatinhd.utils import Unpickler


class Clustering(Flow):
    labels = Stored("labels")
    """Labels for each cell."""

    cluster_info = Stored("cluster_info")
    """Dataframe containing information, such as a label, for each cluster."""

    @classmethod
    def from_labels(cls, labels, path):
        clustering = cls(path)
        if not isinstance(labels, pd.Series):
            labels = pd.Series(labels).astype("category")
        clustering.labels = labels
        clustering.cluster_info = (
            pd.DataFrame(
                {
                    "cluster": labels.unique(),
                    "n_cells": labels.value_counts(),
                    "label": labels.unique(),
                }
            )
            .set_index("cluster")
            .loc[labels.cat.categories]
        )
        return clustering

    @property
    def n_clusters(self):
        return len(self.labels.unique())



from chromatinhd.flow import Flow, Stored, TSV
import typing
import pandas as pd
import numpy as np
import pathlib


class Regions(Flow):
    """
    Regions, typically centered around a transcription start site
    """

    coordinates = TSV("coordinates", columns=["chrom", "start", "end"])
    window = Stored("window")

    @classmethod
    def from_canonical_transcripts(
        cls, canonical_transcripts: pd.DataFrame, window: np.ndarray, path: pathlib.Path
    ):
        regions = canonical_transcripts[
            ["chrom", "start", "end", "ensembl_transcript_id"]
        ].copy()

        regions["tss"] = [
            genes_row["start"] if genes_row["strand"] == +1 else genes_row["end"]
            for _, genes_row in canonical_transcripts.loc[regions.index].iterrows()
        ]
        regions["strand"] = canonical_transcripts["strand"]
        regions["positive_strand"] = (regions["strand"] == 1).astype(int)
        regions["negative_strand"] = (regions["strand"] == -1).astype(int)
        regions["chrom"] = canonical_transcripts.loc[regions.index, "chrom"]

        regions["start"] = (
            regions["tss"]
            + window[0] * (regions["strand"] == 1)
            - window[1] * (regions["strand"] == -1)
        )
        regions["end"] = (
            regions["tss"]
            + window[1] * (regions["strand"] == -1)
            - window[0] * (regions["strand"] == 1)
        )

        return cls.create(
            path=path,
            coordinates=regions[
                ["chrom", "start", "end", "tss", "strand", "ensembl_transcript_id"]
            ],
            window=window,
        )



import numpy as np
import scipy.signal
import pandas as pd


class DifferentialSlices:
    def __init__(
        self, positions, gene_ixs, cluster_ixs, window, n_genes, n_clusters, scores=None
    ):
        assert positions.ndim == 2
        assert positions.shape[1] == 2
        self.positions = positions
        self.gene_ixs = gene_ixs
        self.cluster_ixs = cluster_ixs
        self.window = window
        self.n_genes = n_genes
        self.n_clusters = n_clusters
        self.scores = scores

    def get_slicescores(self):
        slicescores = pd.DataFrame(
            {
                "start": self.positions[:, 0],
                "end": self.positions[:, 1],
                "length": (self.positions[:, 1] - self.positions[:, 0]),
                "gene_ix": self.gene_ixs,
                "cluster_ix": self.cluster_ixs,
                "mid": self.positions[:, 0]
                + (self.positions[:, 1] - self.positions[:, 0]) / 2
                + self.window[0],
            }
        )
        slicescores.index = np.arange(len(slicescores))
        return slicescores

    def get_slicelocations(self, promoters):
        slicelocations = self.get_slicescores().join(
            promoters[["start", "end", "strand", "chr", "gene_ix"]].set_index(
                "gene_ix"
            ),
            on="gene_ix",
            rsuffix="_promoter",
        )

        slicelocations["start_genome"] = np.where(
            slicelocations["strand"] == 1,
            slicelocations["start_promoter"] + slicelocations["start"],
            slicelocations["end_promoter"] - slicelocations["end"],
        )
        slicelocations["end_genome"] = np.where(
            slicelocations["strand"] == 1,
            slicelocations["start_promoter"] + slicelocations["end"],
            slicelocations["end_promoter"] - slicelocations["start"],
        )
        return slicelocations

    def get_randomslicelocations(self, promoters, n_random=10):
        slicescores = self.get_slicescores().copy()
        gene_ixs = np.random.choice(
            len(promoters), len(slicescores) * n_random, replace=True
        )
        slicescores = slicescores.iloc[np.repeat(np.arange(len(slicescores)), n_random)]
        slicescores["gene_ix"] = gene_ixs
        slicelocations = slicescores.join(
            promoters[["start", "end", "strand", "chr", "gene_ix"]].set_index(
                "gene_ix"
            ),
            on="gene_ix",
            rsuffix="_promoter",
        )

        slicelocations["start_genome"] = np.where(
            slicelocations["strand"] == 1,
            slicelocations["start_promoter"] + slicelocations["start"],
            slicelocations["end_promoter"] - slicelocations["end"],
        )
        slicelocations["end_genome"] = np.where(
            slicelocations["strand"] == 1,
            slicelocations["start_promoter"] + slicelocations["end"],
            slicelocations["end_promoter"] - slicelocations["start"],
        )
        return slicelocations

    def get_sliceaverages(self, probs):
        probs_baseline = probs.mean(1)
        slice_average_baseline = []
        slice_max_baseline = []
        slice_std_baseline = []
        slice_average_oi = []
        slice_max_oi = []
        slice_std_oi = []
        slice_average_lfc = []
        slice_max_lfc = []
        slice_summit = []
        for start, end, gene_ix, cluster_ix in zip(
            self.positions[:, 0], self.positions[:, 1], self.gene_ixs, self.cluster_ixs
        ):
            prob_oi = probs[
                gene_ix,
                cluster_ix,
                (start):(end),
            ]

            prob_baseline = probs_baseline[
                gene_ix,
                (start):(end),
            ]
            slice_average_oi.append(prob_oi.mean())
            slice_max_oi.append(prob_oi.max())
            slice_std_oi.append(prob_oi.std())

            slice_average_baseline.append(prob_baseline.mean())
            slice_max_baseline.append(prob_baseline.max())
            slice_std_baseline.append(prob_baseline.std())

            slice_lfc = prob_oi - prob_baseline

            slice_average_lfc.append(slice_lfc.mean())
            slice_max_lfc.append(slice_lfc.max())

            slice_summit.append(start + np.argmax(prob_oi))

        slice_average_baseline = np.hstack(slice_average_baseline)
        slice_max_baseline = np.hstack(slice_max_baseline)
        slice_std_baseline = np.hstack(slice_std_baseline)

        slice_average_oi = np.hstack(slice_average_oi)
        slice_max_oi = np.hstack(slice_max_oi)
        slice_std_oi = np.hstack(slice_std_oi)

        slice_average_lfc = np.hstack(slice_average_lfc)
        slice_max_lfc = np.hstack(slice_max_lfc)

        slice_summit = np.hstack(slice_summit)

        return pd.DataFrame(
            {
                "average_baseline": slice_average_baseline,
                "max_baseline": slice_max_baseline,
                "std_baseline": slice_std_baseline,
                "average": slice_average_oi,
                "max": slice_max_oi,
                "std": slice_std_oi,
                "average_lfc": slice_average_lfc,
                "max_lfc": slice_max_lfc,
                "summit": slice_summit,
            }
        )

    @property
    def position_chosen(self):
        position_chosen = np.zeros(
            (self.n_genes, self.n_clusters, (self.window[1] - self.window[0])),
            # self.n_clusters * self.n_genes * (self.window[1] - self.window[0]),
            dtype=bool,
        )
        for start, end, gene_ix, cluster_ix in zip(
            self.positions[:, 0], self.positions[:, 1], self.gene_ixs, self.cluster_ixs
        ):
            position_chosen[gene_ix, cluster_ix, start:end] = True
        return position_chosen.flatten()

    @property
    def position_ranked(self):
        position_chosen = np.zeros(
            (self.n_genes, self.n_clusters, (self.window[1] - self.window[0])),
            # self.n_clusters * self.n_genes * (self.window[1] - self.window[0]),
        )
        for start, end, gene_ix, cluster_ix, score in zip(
            self.positions[:, 0],
            self.positions[:, 1],
            self.gene_ixs,
            self.cluster_ixs,
            self.scores,
        ):
            position_chosen[gene_ix, cluster_ix, start:end] = score
        return position_chosen

    @property
    def position_indices(self):
        position_chosen = self.position_chosen
        position_indices = np.where(position_chosen)[0]
        return position_indices

    @classmethod
    def from_positions(
        cls,
        positions,
        gene_ixs,
        cluster_ixs,
        window,
        n_genes,
        n_clusters,
        resolution=1,
    ):
        groups = np.hstack(
            [
                0,
                np.cumsum(
                    (
                        np.diff(
                            cluster_ixs * gene_ixs * ((window[1] - window[0]) + 1)
                            + positions
                        )
                        != 1
                    )
                ),
            ]
        )
        cuts = np.where(np.hstack([True, (np.diff(groups) != 0), True]))[0]

        position_slices = (
            np.vstack((positions[cuts[:-1]], positions[cuts[1:] - 1] + 1)).T
            * resolution
        )
        gene_ixs = gene_ixs[cuts[:-1]]
        cluster_ixs = cluster_ixs[cuts[:-1]]
        return cls(position_slices, gene_ixs, cluster_ixs, window, n_genes, n_clusters)

    @classmethod
    def from_peakscores(
        cls,
        peakscores,
        window,
        n_genes,
        logfoldchanges_cutoff=1.0,
        pvals_adj_cutoff=0.05,
    ):

        # get significant
        peakscores["significant"] = (
            peakscores["logfoldchanges"] > logfoldchanges_cutoff
        ) & (peakscores["pvals_adj"] < pvals_adj_cutoff)

        # join peaks that overlap or are adjacent in any way
        # we create an adjacency matrix with the number of shared nucleotides as weight
        # we then check for connected components and merge significant peaks

        import scipy.sparse.csgraph

        peakscores_significant_joined = []
        for (cluster, gene_ix), peakscores_oi in peakscores.query(
            "significant"
        ).groupby(["cluster", "gene_ix"]):
            # only check for overlaps if there are more than 2 signficant peaks
            if len(peakscores_oi) > 1:
                # calculate peak-wise membership within the gene window
                membership = np.zeros((len(peakscores_oi), window[1] - window[0]))
                for i, start, end in zip(
                    np.arange(len(peakscores_oi)),
                    peakscores_oi.relative_start,
                    peakscores_oi.relative_end,
                ):
                    membership[i, (start - window[0]) : ((end - window[0]) + 1)] = 1

                # calculate adjacency
                adjacency = np.matmul(membership, membership.T)

                # get connected components
                connected_components = scipy.sparse.csgraph.connected_components(
                    adjacency, directed=False
                )

                # merge
                peakscores_oi["component"] = connected_components[1]
                peakscores_oi_joined = peakscores_oi.groupby("component").agg(
                    {
                        "relative_start": min,
                        "relative_end": max,
                        "logfoldchanges": "mean",
                    }
                )
                peakscores_oi_joined = peakscores_oi_joined.assign(
                    cluster=cluster, gene_ix=gene_ix
                )
            else:
                peakscores_oi_joined = peakscores_oi.reset_index()[
                    [
                        "gene_ix",
                        "cluster",
                        "relative_start",
                        "relative_end",
                        "logfoldchanges",
                    ]
                ].copy()
            peakscores_significant_joined.append(peakscores_oi_joined)

        peakscores_significant = pd.concat(
            peakscores_significant_joined, ignore_index=True
        )
        peakscores_significant["cluster"] = pd.Categorical(
            peakscores_significant["cluster"],
            categories=peakscores["cluster"].cat.categories,
        )

        # get positions
        positions = peakscores_significant[["relative_start", "relative_end"]].values
        scores = peakscores_significant["logfoldchanges"].values
        positions = positions - window[0]
        gene_ixs = peakscores_significant["gene_ix"].values
        cluster_ixs = peakscores_significant["cluster"].cat.codes

        return cls(
            positions,
            gene_ixs,
            cluster_ixs,
            window,
            n_genes,
            scores=scores,
            n_clusters=len(peakscores["cluster"].cat.categories),
        )

    @classmethod
    def from_basepair_ranking(cls, basepair_ranking, window, cutoff, resolution=1):
        """
        :param: cutoff

        """
        n_genes = basepair_ranking.shape[0]
        n_clusters = basepair_ranking.shape[1]

        basepairs_oi = basepair_ranking > cutoff

        gene_ixs, cluster_ixs, positions = np.where(basepairs_oi)

        return cls.from_positions(
            positions,
            gene_ixs,
            cluster_ixs,
            window,
            n_genes,
            n_clusters,
            resolution=resolution,
        )

    def get_slicetopologies(self, probs):
        probs_mean = probs.mean(1)
        import itertools

        prominences = []
        n_subpeaks = []
        balances = []
        balances_raw = []
        dominances = []
        shadows = []
        differentialdominances = []
        for gene_ix, cluster_ix in itertools.product(
            range(self.n_genes), range(self.n_clusters)
        ):
            slices_oi = (self.gene_ixs == gene_ix) & (self.cluster_ixs == cluster_ix)
            position_slices_gene = self.positions[slices_oi]
            probs_gene = probs[gene_ix, cluster_ix]
            probs_mean_gene = probs_mean[gene_ix]
            x = np.exp(probs_gene)
            peaks = []
            for slice in position_slices_gene:
                peaks.append(slice[0] + np.argmax(x[slice[0] : (slice[1] + 1)]))

                # determine number of supbeaks
                height_threshold = 0.25
                distance_threshold = 80
                x_ = x[slice[0] : (slice[1] + 1)]
                y_ = np.exp(probs_mean_gene[slice[0] : (slice[1] + 1)])
                subpeaks, subpeaks_info = scipy.signal.find_peaks(
                    x_, height=x_.max() * height_threshold, distance=distance_threshold
                )
                n_subpeaks.append(len(subpeaks))

                # determine local neighbourhood
                distance = 800

                neighborhoud_left = [
                    np.clip(slice[0] - distance, *(self.window - self.window[0])),
                    slice[0],
                ]
                neighborhoud_right = [
                    slice[1],
                    np.clip(slice[1] + distance, *(self.window - self.window[0])),
                ]

                x_neighborhoud = np.exp(
                    np.hstack(
                        [
                            probs_gene[
                                neighborhoud_left[0] : (neighborhoud_left[1] + 1)
                            ],
                            probs_gene[
                                neighborhoud_right[0] : (neighborhoud_right[1] + 1),
                            ],
                        ]
                    )
                )
                y_neighborhoud = np.exp(
                    np.hstack(
                        [
                            probs_mean_gene[
                                neighborhoud_left[0] : (neighborhoud_left[1] + 1)
                            ],
                            probs_mean_gene[
                                neighborhoud_right[0] : (neighborhoud_right[1] + 1)
                            ],
                        ]
                    )
                )

                dominances.append((x_.max() / max(x_.max(), x_neighborhoud.max())))

                differentialdominances.append(
                    (x_ / y_).max()
                    / max((x_ / y_).max(), (x_neighborhoud / y_neighborhoud).max())
                )

                # determine local "balancing"
                balances.append(
                    np.clip(y_neighborhoud - x_neighborhoud, 0, np.inf).sum()
                    / (x_ - y_).sum()
                )
                balances_raw.append((x_neighborhoud / y_neighborhoud).min())

                # determine local shadow
                shadows.append(x_.mean() / x_.max())

            peaks = np.array(peaks, dtype=int)

            # calculate prominence
            # we have to pad the left and right side because peaks in the margins do not get the correct prominence
            prominence, left_bases, right_bases = scipy.signal.peak_prominences(
                np.hstack([0, x, 0]), peaks + 1
            )
            relative_prominence = prominence / x[peaks]
            prominences.append(relative_prominence)

        prominences = np.hstack(prominences)
        shadows = np.hstack(shadows)

        return pd.DataFrame(
            {
                "prominence": prominences,
                "n_subpeaks": n_subpeaks,
                "log1p_n_subpeaks": n_subpeaks,
                "balances_raw": balances_raw,
                "balance": balances,
                "dominance": dominances,
                "differentialdominance": differentialdominances,
                "shadow": shadows,
            }
        )

    def get_slicedifferential(self, probs):
        probs_diff = probs - probs.mean(1, keepdims=True)
        import itertools

        slice_differential = []
        for start, end, gene_ix, cluster_ix in zip(
            self.positions[:, 0], self.positions[:, 1], self.gene_ixs, self.cluster_ixs
        ):
            probs_diff_oi = probs_diff[
                gene_ix,
                cluster_ix,
                (start):(end),
            ]
            slice_differential.append((probs_diff_oi > np.log(2)).mean())
        slice_differential = np.hstack(slice_differential)

        return pd.DataFrame(
            {
                "differential_positions": slice_differential,
            }
        )



import tqdm.auto as tqdm
import torch
import numpy as np
from chromatinhd.train import Trace

import logging

logger = logging.getLogger(__name__)
logger.setLevel(logging.WARNING)


def paircor(x, y, dim=0, eps=0.1):
    divisor = (y.std(dim) * x.std(dim)) + eps
    cor = ((x - x.mean(dim, keepdims=True)) * (y - y.mean(dim, keepdims=True))).mean(
        dim
    ) / divisor
    return cor


def filter_minibatch_sets(minibatch_sets, improved):
    new_minibatch_sets = []
    for minibatch_set in minibatch_sets:
        tasks = [
            minibatch.filter_genes(improved) for minibatch in minibatch_set["tasks"]
        ]
        tasks = [minibatch for minibatch in tasks if len(minibatch.genes_oi) > 0]
        new_minibatch_sets.append({"tasks": tasks})
    return new_minibatch_sets


class Trainer:
    def __init__(
        self,
        model,
        loaders_train,
        loaders_validation,
        minibatcher_train,
        minibatcher_validation,
        optim,
        device="cuda",
        n_epochs=30,
        checkpoint_every_epoch=1,
        optimize_every_step=10,
    ):
        self.model = model
        self.loaders_train = loaders_train
        self.loaders_validation = loaders_validation

        self.trace = Trace()

        self.optim = optim

        self.step_ix = 0
        self.epoch = 0
        self.n_epochs = n_epochs

        self.checkpoint_every_epoch = checkpoint_every_epoch
        self.optimize_every_step = optimize_every_step

        self.minibatcher_train = minibatcher_train
        self.minibatcher_validation = minibatcher_validation

        self.device = device

    def train(self):
        self.model = self.model.to(self.device)

        continue_training = True

        prev_gene_loss = None
        improved = None

        self.loaders_train.initialize(self.minibatcher_train)
        self.loaders_validation.initialize(self.minibatcher_validation)

        n_steps_total = self.n_epochs * len(self.loaders_train)
        pbar = tqdm.tqdm(total=n_steps_total, leave=False)

        while (self.epoch < self.n_epochs) and (continue_training):
            # checkpoint if necessary
            if (self.epoch % self.checkpoint_every_epoch) == 0:
                with torch.no_grad():
                    gene_loss = np.zeros(self.minibatcher_train.n_genes)
                    for data_validation in self.loaders_validation:
                        data_validation = data_validation.to(self.device)

                        gene_loss_mb = (
                            self.model.forward_gene_loss(data_validation)
                            .cpu()
                            .detach()
                            .numpy()
                        )

                        gene_loss[data_validation.minibatch.genes_oi] = (
                            gene_loss[data_validation.minibatch.genes_oi] + gene_loss_mb
                        )

                self.trace.append(
                    gene_loss.mean().item(),
                    self.epoch,
                    self.step_ix,
                    "validation",
                )
                logger.info(f"{'•'} {self.epoch}/{self.n_epochs} {'step':>15}")
                self.trace.checkpoint(logger=logger)

                # compare with previous loss per gene
                if prev_gene_loss is not None:
                    improvement = gene_loss - prev_gene_loss

                    if improved is not None:
                        improved = improved & (improvement < 0)
                    else:
                        improved = improvement < 0
                    logger.info(f"{improved.mean():.1%}")

                    # stop training once less than 1% of genes are still being optimized
                    if improved.mean() < 0.001:
                        break

                    self.minibatcher_train.genes = np.arange(
                        self.minibatcher_train.n_genes
                    )[improved]

                prev_gene_loss = gene_loss.copy()

            # train
            for data_train in self.loaders_train:
                data_train = data_train.to(self.device)

                loss = self.model.forward_loss(data_train)
                loss.backward()

                # check if optimization
                if (self.step_ix % self.optimize_every_step) == 0:
                    self.optim.step()
                    self.optim.zero_grad()

                self.step_ix += 1
                pbar.update()

                self.trace.append(loss.item(), self.epoch, self.step_ix, "train")
            self.epoch += 1

        pbar.update(n_steps_total)
        pbar.close()

        self.model = self.model.to("cpu")



import chromatinhd as chd
import pytest

import pathlib


@pytest.fixture(scope="session")
def example_dataset_folder(tmp_path_factory):
    example_dataset_folder = tmp_path_factory.mktemp("example")

    import pkg_resources
    import shutil

    DATA_PATH = pathlib.Path(
        pkg_resources.resource_filename("chromatinhd", "data/examples/pbmc10ktiny/")
    )

    # copy all files from data path to dataset folder
    for file in DATA_PATH.iterdir():
        shutil.copy(file, example_dataset_folder / file.name)
    return example_dataset_folder


@pytest.fixture(scope="session")
def example_transcriptome(example_dataset_folder):
    import scanpy as sc

    adata = sc.read(example_dataset_folder / "transcriptome.h5ad")
    transcriptome = chd.data.Transcriptome.from_adata(
        adata, path=example_dataset_folder / "transcriptome"
    )
    return transcriptome


@pytest.fixture(scope="session")
def example_clustering(example_dataset_folder, example_transcriptome):
    clustering = chd.data.Clustering.from_labels(
        example_transcriptome.adata.obs["celltype"],
        path=example_dataset_folder / "clustering",
    )
    return clustering


@pytest.fixture(scope="session")
def example_regions(example_dataset_folder, example_transcriptome):
    biomart_dataset = chd.biomart.Dataset.from_genome("GRCh38")
    canonical_transcripts = chd.biomart.get_canonical_transcripts(
        biomart_dataset, example_transcriptome.var.index
    )
    regions = chd.data.Regions.from_canonical_transcripts(
        canonical_transcripts,
        path=example_dataset_folder / "regions",
        window=[-10000, 10000],
    )
    return regions


@pytest.fixture(scope="session")
def example_fragments(example_dataset_folder, example_transcriptome, example_regions):
    fragments = chd.data.Fragments.from_fragments_tsv(
        example_dataset_folder / "fragments.tsv.gz",
        example_regions,
        obs=example_transcriptome.obs,
        path=example_dataset_folder / "fragments",
    )
    fragments.create_cellxgene_indptr()
    return fragments



import torch
import numpy as np
import dataclasses

@dataclasses.dataclass
class Result():
    counts:np.ndarray
    cells_oi:np.ndarray
    genes_oi:np.ndarray
    @property
    def n_cells(self):
        return len(self.cells_oi)

    @property
    def n_genes(self):
        return len(self.genes_oi)

    def to(self, device):
        for field_name, field in self.__dataclass_fields__.items():
            if field.type is torch.Tensor:
                self.__setattr__(field_name, self.__getattribute__(field_name).to(device))
        return self

    @property
    def genes_oi_torch(self):
        return torch.from_numpy(self.genes_oi).to(self.coordinates.device)

    @property
    def cells_oi_torch(self):
        return torch.from_numpy(self.cells_oi).to(self.coordinates.device)

class PeakcountsResult(Result):
    pass

class Peakcounts():
    def __init__(self, fragments, peakcounts):
        self.peakcounts = peakcounts
        assert "gene_ix" in peakcounts.peaks.columns
        var = peakcounts.var
        var["ix"] = np.arange(peakcounts.var.shape[0])
        peakcounts.var = var
        assert "ix" in peakcounts.var.columns

        assert peakcounts.counts.shape[1] == peakcounts.var.shape[0]

        self.gene_peak_mapping = []
        i = 0
        cur_gene_ix = -1
        for peak_ix, gene_ix in zip(peakcounts.var["ix"][peakcounts.peaks.index].values, peakcounts.peaks["gene_ix"]):
            while gene_ix != cur_gene_ix:
                self.gene_peak_mapping.append([])
                cur_gene_ix += 1
            self.gene_peak_mapping[-1].append(peak_ix)

        
    def load(self, minibatch):
        peak_ixs = np.concatenate([self.gene_peak_mapping[gene_ix] for gene_ix in minibatch.genes_oi])
        counts = self.peakcounts.counts[minibatch.cells_oi, :][:, peak_ixs]
        
        return PeakcountsResult(
            counts = counts,
            **minibatch.items()
        )



import numpy as np
import pandas as pd


def digits2number(digits, base=2, fractional=False):
    powers = np.arange(len(digits))
    out = (digits * base**powers).sum()
    if fractional:
        out = out / base ** (len(digits))
    return out


def number2digits(n, base=2):
    if n == 0:
        return np.array()
    nDigits = np.ceil(np.log(n + 1) / np.log(base))
    powers = base ** (np.arange(nDigits + 1))
    out = np.diff(n % powers) / powers[:-1]
    return out


def vanDerCorput(n, base=2, start=1):
    out = np.array(
        [
            digits2number(number2digits(ii, base)[::-1], base, True)
            for ii in range(1, n + start)
        ]
    )
    return out


def offset(
    y,
    maxLength=None,
    method="quasirandom",
    nbins=20,
    adjust=1,
    bw_method="scott",
    max_density=None,
):
    if len(y) == 1:
        return [0]

    if isinstance(y, pd.Series):
        y = y.values

    if nbins is None:
        if method in ["pseudorandom", "quasirandom"]:
            nbins = 2**10
        else:
            nbins = int(max(2, np.ceil(len(y) / 5)))

    if maxLength is None:
        subgroup_width = 1
    else:
        subgroup_width = np.sqrt(len(y) / maxLength)

    #     from sklearn.neighbors import KernelDensity
    #     kde = KernelDensity(kernel='gaussian').fit(y.reshape(-1, 1))
    #     pointDensities = np.exp(np.array(kde.score_samples(y.reshape(-1, 1))))

    import scipy.stats

    kernel = scipy.stats.gaussian_kde(y, bw_method=bw_method)
    pointDensities = kernel(y)

    if max_density is None:
        max_density = pointDensities.max()

    pointDensities = pointDensities / max_density

    if method == "quasirandom":
        offset = np.array(vanDerCorput(len(y)))[np.argsort(y)]
    elif method == "pseudorandom":
        offset = np.random.uniform(size=len(y))

    out = (offset - 0.5) * 2 * pointDensities * subgroup_width * 0.5

    return out



import sys
import numpy as np
import pyximport

pyximport.install(
    reload_support=True,
    language_level=3,
    setup_args=dict(include_dirs=[np.get_include()]),
)
import chromatinhd.loaders.extraction.fragments
import chromatinhd.loaders.extraction.motifs

import torch
from multiprocessing import shared_memory

import dataclasses

from .fragments import Fragments

from .fragments import Result


@dataclasses.dataclass
class FullResult(Result):
    motifcounts: torch.Tensor
    n_motifs: int


class Full(Fragments):
    def __init__(self, fragments, motifscan, cellxgene_batch_size, window, cutwindow):
        super().__init__(fragments, cellxgene_batch_size, window)

        # store auxilliary information
        self.cutwindow = cutwindow
        self.cutwindow_width = cutwindow[1] - cutwindow[0]

        # create buffers for motifs
        n_motifs = motifscan.shape[1]
        n_motifs_per_fragment = 1000  # 400 motifs
        motif_buffer_size = self.fragment_buffer_size * n_motifs_per_fragment

        self.motifscan_indptr = motifscan.indptr.astype(np.int64)
        self.motifscan_indices = motifscan.indices.astype(np.int64)
        self.motifscan_data = motifscan.data.astype(np.float64)

        self.out_fragment_indptr = torch.from_numpy(
            np.zeros(motif_buffer_size, dtype=int)
        )  # .pin_memory()
        self.out_motif_ix = torch.from_numpy(
            np.zeros(motif_buffer_size, dtype=int)
        )  # .pin_memory()
        self.out_score = torch.from_numpy(
            np.zeros(motif_buffer_size, dtype=np.float64)
        )  # .pin_memory()
        self.out_distance = torch.from_numpy(
            np.zeros(motif_buffer_size, dtype=int)
        )  # .pin_memory()
        self.out_motifcounts = torch.from_numpy(
            np.ascontiguousarray(
                np.zeros((self.fragment_buffer_size, n_motifs), dtype=int)
            )
        )  # .pin_memory()

    def load(self, minibatch, **kwargs):
        super().load(minibatch, **kwargs)

        n_fragments = self.out_coordinates.shape[0]

        self.out_motifcounts.data.zero_()  # this is a big slow down (~20% of function cpu time) but unsure how to fix

        n_motifs = chromatinhd.loaders.extraction.motifs.extract_all(
            self.out_coordinates.numpy(),
            self.out_genemapping.numpy(),
            self.motifscan_indptr,
            self.motifscan_indices,
            self.motifscan_data,
            *self.window,
            self.window_width,
            *self.cutwindow,
            self.out_fragment_indptr.numpy(),
            self.out_motif_ix.numpy(),
            self.out_score.numpy(),
            self.out_distance.numpy(),
            self.out_motifcounts.numpy()
        )
        self.out_fragment_indptr.resize_(n_fragments + 1)
        self.out_motif_ix.resize_(n_motifs)
        self.out_score.resize_(n_motifs)
        self.out_distance.resize_(n_motifs)
        self.out_motifcounts.resize_((n_fragments, self.out_motifcounts.shape[1]))

        return FullResult(
            motifcounts=self.out_motifcounts,
            local_cellxgene_ix=self.out_local_cellxgene_ix,
            n_fragments=n_fragments,
            **minibatch.items()
        )


@dataclasses.dataclass
class MotifcountsResult(Result):
    motifcounts: torch.Tensor


class Motifcounts(Fragments):
    def __init__(
        self, fragments, motifscan, cellxgene_batch_size, window, cutwindow, **kwargs
    ):
        super().__init__(fragments, cellxgene_batch_size, window, **kwargs)

        # store auxilliary information
        self.cutwindow = cutwindow
        self.cutwindow_width = cutwindow[1] - cutwindow[0]

        # create buffers for motifs
        self.n_motifs = motifscan.shape[1]
        self.n_features = self.n_motifs

        self.motifscan_indptr = motifscan.indptr.astype(np.int64)
        self.motifscan_indices = motifscan.indices.astype(np.int64)
        self.motifscan_data = motifscan.data.astype(np.float64)

    def load(self, minibatch, **kwargs):
        super().load(minibatch, **kwargs)

        n_fragments = self.out_coordinates.shape[0]

        out_motifcounts = np.zeros((n_fragments, self.n_features), dtype=np.int64)

        n_motifs = chromatinhd.loaders.extraction.motifs.extract_motifcounts(
            self.out_coordinates.numpy(),
            self.out_genemapping.numpy(),
            self.motifscan_indptr,
            self.motifscan_indices,
            self.motifscan_data,
            *self.window,
            self.window_width,
            *self.cutwindow,
            out_motifcounts
        )
        out_motifcounts.resize((n_fragments, self.n_features))

        return MotifcountsResult(
            motifcounts=torch.from_numpy(out_motifcounts).to(torch.float),
            local_cellxgene_ix=self.out_local_cellxgene_ix,
            coordinates=self.out_coordinates,
            genemapping=self.out_genemapping,
            n_fragments=n_fragments,
            **minibatch.items()
        )


class MotifcountsSplit(Fragments):
    def __init__(
        self, fragments, motifscan, cellxgene_batch_size, window, cutwindow, **kwargs
    ):
        super().__init__(fragments, cellxgene_batch_size, window, **kwargs)

        # store auxilliary information
        self.cutwindow = cutwindow
        self.cutwindow_width = cutwindow[1] - cutwindow[0]

        # create buffers for motifs
        self.n_motifs = motifscan.shape[1]
        self.n_features = self.n_motifs * 2

        self.motifscan_indptr = motifscan.indptr
        self.motifscan_indices = motifscan.indices
        self.motifscan_data = motifscan.data

    def load(self, minibatch, **kwargs):
        super().load(minibatch, **kwargs)

        n_fragments = self.out_coordinates.shape[0]

        out_motifcounts = np.zeros((n_fragments, self.n_features), dtype=np.int64)

        n_motifs = chromatinhd.loaders.extraction.motifs.extract_motifcounts_split(
            self.out_coordinates.numpy(),
            self.out_genemapping.numpy(),
            self.motifscan_indptr,
            self.motifscan_indices,
            self.motifscan_data,
            self.n_motifs,
            *self.window,
            self.window_width,
            *self.cutwindow,
            out_motifcounts
        )
        out_motifcounts.resize((n_fragments, self.n_features))

        return MotifcountsResult(
            motifcounts=torch.from_numpy(out_motifcounts).to(torch.float),
            local_cellxgene_ix=self.out_local_cellxgene_ix,
            coordinates=self.out_coordinates,
            genemapping=self.out_genemapping,
            n_fragments=n_fragments,
            **minibatch.items()
        )


class MotifcountsMultiple(Fragments):
    def __init__(
        self, fragments, motifscan, cellxgene_batch_size, window, cutwindows, **kwargs
    ):
        super().__init__(fragments, cellxgene_batch_size, window, **kwargs)

        # store auxilliary information
        self.cutwindows = cutwindows

        # create buffers for motifs
        self.n_motifs = motifscan.shape[1]
        self.n_features = self.n_motifs * (cutwindows.shape[0] - 1)

        self.motifscan_indptr = motifscan.indptr
        self.motifscan_indices = motifscan.indices
        self.motifscan_data = motifscan.data

    def load(self, minibatch, **kwargs):
        super().load(minibatch, **kwargs)

        n_fragments = self.out_coordinates.shape[0]

        out_motifcounts = np.zeros((n_fragments, self.n_features), dtype=np.int64)

        n_motifs = chromatinhd.loaders.extraction.motifs.extract_motifcounts_multiple(
            self.out_coordinates.numpy(),
            self.out_genemapping.numpy(),
            self.motifscan_indptr,
            self.motifscan_indices,
            self.motifscan_data,
            self.n_motifs,
            *self.window,
            self.window_width,
            self.cutwindows,
            out_motifcounts
        )
        out_motifcounts.resize((n_fragments, self.n_features))

        return MotifcountsResult(
            motifcounts=torch.from_numpy(out_motifcounts).to(torch.float),
            local_cellxgene_ix=self.out_local_cellxgene_ix,
            coordinates=self.out_coordinates,
            genemapping=self.out_genemapping,
            n_fragments=n_fragments,
            **minibatch.items()
        )


class MotifcountsRelative(Fragments):
    def __init__(
        self,
        fragments,
        motifscan,
        cellxgene_batch_size,
        window,
        cutwindow,
        promoter_width,
        **kwargs
    ):
        super().__init__(fragments, cellxgene_batch_size, window, **kwargs)

        # store auxilliary information
        self.cutwindow = cutwindow
        self.cutwindow_width = cutwindow[1] - cutwindow[0]

        # create buffers for motifs
        self.n_motifs = motifscan.shape[1]
        self.n_features = self.n_motifs * 2

        self.promoter_width = promoter_width

        self.motifscan_indptr = motifscan.indptr.astype(np.int64)
        self.motifscan_indices = motifscan.indices.astype(np.int64)
        self.motifscan_data = motifscan.data.astype(np.float64)

    def load(self, minibatch, **kwargs):
        super().load(minibatch, **kwargs)

        n_fragments = self.out_coordinates.shape[0]

        out_motifcounts = np.zeros((n_fragments, self.n_features), dtype=np.int64)

        n_motifs = chromatinhd.loaders.extraction.motifs.extract_motifcounts_relative(
            self.out_coordinates.numpy(),
            self.out_genemapping.numpy(),
            self.motifscan_indptr,
            self.motifscan_indices,
            self.motifscan_data,
            self.n_motifs,
            *self.window,
            self.window_width,
            *self.cutwindow,
            self.promoter_width,
            out_motifcounts
        )
        out_motifcounts.resize((n_fragments, self.n_features))

        return MotifcountsResult(
            motifcounts=torch.from_numpy(out_motifcounts).to(torch.float),
            local_cellxgene_ix=self.out_local_cellxgene_ix,
            coordinates=self.out_coordinates,
            genemapping=self.out_genemapping,
            n_fragments=n_fragments,
            **minibatch.items()
        )



import pandas as pd
import numpy as np


phases = pd.DataFrame(
    {"phase": ["train", "validation"], "color": ["#888888", "tomato"]}
).set_index("phase")


colors = ["#0074D9", "#FF4136", "#FF851B", "#2ECC40", "#39CCCC", "#85144b"]


def replace_patch(ax, patch, points=10, ha="center", va="center"):
    """
    Replaces a patch, often an axis label, with a new rectangular axis according to figure size "points". Eample usage:


    fig, ax = plt.subplots()
    ax.set_title("100%, (0.5,1-0.3,.3,.3)")
    x.plot([0, 2], [0, 2])

    for l in ax.get_xticklabels():
        ax1 = replace_patch(ax, l)
        ax1.plot([0, 1], [0, 1])
    """
    fig = ax.get_figure()

    fig.draw_without_rendering()

    w, h = fig.transFigure.inverted().transform([[1, 1]]).ravel() * points
    bbox = fig.transFigure.inverted().transform(patch.get_window_extent())
    dw = bbox[1, 0] - bbox[0, 0]
    dh = bbox[1, 1] - bbox[0, 1]
    x = bbox[0, 0] + dw / 2
    y = bbox[0, 1] + dh / 2

    if ha == "left":
        x += 0
    elif ha == "right":
        x -= w - dw / 2
    elif ha == "center":
        x -= w / 2

    if va == "bottom":
        y += 0
    elif va == "top":
        y -= h - dh / 2
    elif va == "center":
        y -= h / 2

    ax1 = fig.add_axes([x, y, w, h])
    # ax1.axis("off")

    return ax1



# %%
import requests

# Send a GET request to the BioMart REST API
response = requests.get('http://www.ensembl.org/biomart/martservice?type=registry')

# The response is an XML string, so parse it using ElementTree
import xml.etree.ElementTree as ET
root = ET.fromstring(response.text)

# Iterate over all the MartURLLocation elements (these represent the datasets)
for dataset in root.iter('MartURLLocation'):
    # The species is stored in the 'name' attribute
    species = dataset.get('name')

    # If the species is human, print the details of the dataset
    if 'hsapiens' in species:
        print(ET.tostring(dataset, encoding='utf8').decode('utf8'))

# %%
response.text
# %%
import pandas as pd
import io

def get_datasets():
    mart = "ENSEMBL_MART_ENSEMBL"
    baseurl = "http://www.ensembl.org/biomart/martservice?"
    url = "{baseurl}type=datasets&requestid=biomaRt&mart={mart}"
    response = requests.get(url.format(baseurl=baseurl, mart=mart))
    root = pd.read_table(io.StringIO(response.text), sep="\t", header=None, names=["_", "dataset", "description", "version", "assembly", "__", "___", "____", "last_update"])
    print(root)
# %%
get_biomart_datasets()
# %%



import torch

def normlinreg(x, y):
    assert x.ndim == 2
    assert x.shape == y.shape
    X = torch.stack([torch.ones(x.shape, device = x.device), x], 1)
    
    coefficients = (torch.linalg.inv(X @ X.transpose(-1, -2)) @ X) @ y.unsqueeze(-1)
    
    return x * coefficients[:, 1] + coefficients[:, 0]

class MSENormLoss2(torch.nn.MSELoss):
    def forward(self, input: torch.Tensor, target: torch.Tensor):
        input_normalized = normlinreg(input, target)
        return super().forward(input_normalized, target) * 100

class MSENormLoss(torch.nn.MSELoss):
    def __init__(self, *args, dim = 0, eps = 1e-1, **kwargs):
        self.dim = dim
        self.eps = eps
        super().__init__(*args, **kwargs)
    def forward(self, input: torch.Tensor, target: torch.Tensor):
        input_normalized = (
            (input - input.mean(self.dim, keepdim = True) + target.mean(self.dim, keepdim = True)) * 
            (target.std(self.dim, keepdim = True) / (input.std(self.dim, keepdim = True) + self.eps))
        )
        return super().forward(input_normalized, target) * 100


class TestModel:
    def test_example(self, example_fragments, example_clustering):
        pass
        # fragments = []
        # clustering = []
        # fold = []
        # model = []
        # result = model.get_prediction(
        #     fragments, clustering, cell_ixs=fold["cells_test"]
        # )



#define NPY_NO_DEPRECATED_API NPY_1_7_API_VERSION
#cython: language_level=3
cimport cython
import numpy as np
cimport numpy as np
np.import_array()

INT64 = np.int64
ctypedef np.int64_t INT64_t
FLOAT64 = np.float64
ctypedef np.float64_t FLOAT64_t

@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.cdivision
def extract_fragments(
    INT64_t [::1] cellxgene_oi,
    INT64_t [::1] cellxgene_indptr,
    INT64_t [:,::1] coordinates,
    INT64_t [::1] genemapping,
    INT64_t [:,::1] out_coordinates,
    INT64_t [::1] out_genemapping,
    INT64_t [::1] out_local_cellxgene_ix,
):
    cdef INT64_t out_ix, local_cellxgene_ix, cellxgene_ix, position
    out_ix = 0 # will store where in the output array we are currently
    local_cellxgene_ix = 0 # will store the current fragment counting from 0

    with nogil:
        for local_cellxgene_ix in range(cellxgene_oi.shape[0]):    
            cellxgene_ix = cellxgene_oi[local_cellxgene_ix]
            for position in range(cellxgene_indptr[cellxgene_ix], cellxgene_indptr[cellxgene_ix+1]):
                out_coordinates[out_ix] = coordinates[position]
                out_genemapping[out_ix] = genemapping[position]
                out_local_cellxgene_ix[out_ix] = local_cellxgene_ix

                out_ix += 1

    return out_ix


import numpy as np
import pandas as pd
import pickle

from chromatinhd.flow import Flow, Stored, StoredDict
from chromatinhd import sparse
from chromatinhd.utils import Unpickler


class Transcriptome(Flow):
    _var = None

    @property
    def var(self):
        if self._var is None:
            self._var = pd.read_table(self.path / "var.tsv", index_col=0)
        return self._var

    @var.setter
    def var(self, value):
        value.index.name = "gene"
        value.to_csv(self.path / "var.tsv", sep="\t")
        self._var = value

    _obs = None

    @property
    def obs(self):
        if self._obs is None:
            self._obs = pd.read_table(self.path / "obs.tsv", index_col=0)
        return self._obs

    @obs.setter
    def obs(self, value):
        value.index.name = "cell"
        value.to_csv(self.path / "obs.tsv", sep="\t")
        self._obs = value

    _adata = None

    @property
    def adata(self):
        if self._adata is None:
            self._adata = pickle.load((self.path / "adata.pkl").open("rb"))
        return self._adata

    @adata.setter
    def adata(self, value):
        pickle.dump(value, (self.path / "adata.pkl").open("wb"))
        self._adata = value

    def gene_id(self, symbol):
        assert all(pd.Series(symbol).isin(self.var["symbol"])), set(
            pd.Series(symbol)[~pd.Series(symbol).isin(self.var["symbol"])]
        )
        return self.var.reset_index("gene").set_index("symbol").loc[symbol]["gene"]

    def symbol(self, gene_id):
        assert all(pd.Series(gene_id).isin(self.var.index)), set(
            pd.Series(gene_id)[~pd.Series(gene_id).isin(self.var.index)]
        )
        return self.var.loc[gene_id]["symbol"]

    def gene_ix(self, symbol):
        self.var["ix"] = np.arange(self.var.shape[0])
        assert all(pd.Series(symbol).isin(self.var["symbol"])), set(
            pd.Series(symbol)[~pd.Series(symbol).isin(self.var["symbol"])]
        )
        return self.var.reset_index("gene").set_index("symbol").loc[symbol]["ix"]

    def create_X(self):
        X_scipy = self.adata.X
        if isinstance(X_scipy, np.ndarray):
            import scipy.sparse

            X_scipy = scipy.sparse.csr_matrix(X_scipy)
        X = sparse.COOMatrix.from_scipy_csr(X_scipy)
        X.populate_mapping()

        self.X = X

    _X = None

    @property
    def X(self):
        if self._X is None:
            self._X = Unpickler((self.path / "X.pkl").open("rb")).load()
        return self._X

    @X.setter
    def X(self, value):
        pickle.dump(value, (self.path / "X.pkl").open("wb"))
        self._X = value

    @classmethod
    def from_adata(cls, adata, path):
        transcriptome = cls(path=path)
        transcriptome.adata = adata
        transcriptome.layers["X"] = adata.X
        transcriptome.var = adata.var
        transcriptome.obs = adata.obs
        return transcriptome

    layers = StoredDict("layers", Stored)


class ClusterTranscriptome(Flow):
    var = Stored("var")
    obs = Stored("obs")
    adata = Stored("adata")
    X = Stored("X")

    def gene_id(self, symbol):
        assert all(pd.Series(symbol).isin(self.var["symbol"])), set(
            pd.Series(symbol)[~pd.Series(symbol).isin(self.var["symbol"])]
        )
        return self.var.reset_index("gene").set_index("symbol").loc[symbol]["gene"]

    def symbol(self, gene_id):
        assert all(pd.Series(gene_id).isin(self.var.index)), set(
            pd.Series(gene_id)[~pd.Series(gene_id).isin(self.var.index)]
        )
        return self.var.loc[gene_id]["symbol"]

    def gene_ix(self, symbol):
        self.var["ix"] = np.arange(self.var.shape[0])
        assert all(pd.Series(symbol).isin(self.var["symbol"])), set(
            pd.Series(symbol)[~pd.Series(symbol).isin(self.var["symbol"])]
        )
        return self.var.reset_index("gene").set_index("symbol").loc[symbol]["ix"]


class ClusteredTranscriptome(Flow):
    donors_info = Stored("donors_info")
    clusters_info = Stored("clusters_info")
    var = Stored("var")
    X = Stored("X")

    def gene_id(self, symbol):
        assert all(pd.Series(symbol).isin(self.var["symbol"])), set(
            pd.Series(symbol)[~pd.Series(symbol).isin(self.var["symbol"])]
        )
        return self.var.reset_index("gene").set_index("symbol").loc[symbol]["gene"]

    def symbol(self, gene_id):
        assert all(pd.Series(gene_id).isin(self.var.index)), set(
            pd.Series(gene_id)[~pd.Series(gene_id).isin(self.var.index)]
        )
        return self.var.loc[gene_id]["symbol"]

    def gene_ix(self, symbol):
        self.var["ix"] = np.arange(self.var.shape[0])
        assert all(pd.Series(symbol).isin(self.var["symbol"])), set(
            pd.Series(symbol)[~pd.Series(symbol).isin(self.var["symbol"])]
        )
        return self.var.reset_index("gene").set_index("symbol").loc[symbol]["ix"]



import chromatinhd.data.fragments
import dataclasses

from .clustering import Clustering
from .minibatches import Minibatch
from .cuts import Cuts


@dataclasses.dataclass
class Result:
    cuts: Cuts
    clustering: Clustering
    minibatch: Minibatch

    def to(self, device):
        self.cuts.to(device)
        self.clustering.to(device)
        self.minibatch.to(device)
        return self


class ClusteringCuts:
    """
    Provides both clustering and cuts data for a minibatch.
    """

    def __init__(
        self,
        clustering: chromatinhd.data.clustering.Clustering,
        fragments: chromatinhd.data.fragments.Fragments,
        cellxgene_batch_size: int,
    ):
        self.clustering = Clustering(clustering)
        self.cuts = Cuts(fragments, cellxgene_batch_size=cellxgene_batch_size)

    def load(self, minibatch):
        return Result(
            cuts=self.cuts.load(minibatch),
            clustering=self.clustering.load(minibatch),
            minibatch=minibatch,
        )



import matplotlib as mpl
import seaborn as sns
import pandas as pd
import numpy as np
import chromatinhd
import scanpy as sc


def get_cmap_rna_diff():
    return mpl.cm.BuGn


class DifferentialExpression(chromatinhd.grid.Wrap):
    def __init__(
        self,
        plotdata_expression_clusters,
        cluster_info,
        width=0.5,
        panel_height=0.5,
        norm_expression=None,
        symbol=None,
        **kwargs,
    ):
        super().__init__(ncol=1, **{"padding_height": 0, **kwargs})

        plotdata_expression_clusters = plotdata_expression_clusters.loc[
            ~plotdata_expression_clusters.index.isin(["Plasma"])
        ]

        if norm_expression is None:
            norm_expression = mpl.colors.Normalize(
                0.0, plotdata_expression_clusters.max(), clip=True
            )

        cmap_expression = get_cmap_rna_diff()

        for cluster_id in cluster_info.index:
            panel, ax = chromatinhd.grid.Ax((width, panel_height))
            self.add(panel)

            sns.despine(ax=ax, left=True, right=True, top=True, bottom=True)
            ax.set_yticks([])
            ax.set_xticks([])

            circle = mpl.patches.Circle(
                (0, 0),
                norm_expression(plotdata_expression_clusters[cluster_id]) * 0.9 + 0.1,
                fc=cmap_expression(
                    norm_expression(plotdata_expression_clusters[cluster_id])
                ),
                lw=1,
                ec="#333333",
            )
            ax.add_patch(circle)
            ax.set_xlim(-1.05, 1.05)
            ax.set_ylim(-1.05, 1.05)
            ax.set_aspect(1)
            text = ax.text(
                1.05,
                0,
                cluster_info.loc[cluster_id, "label"],
                ha="left",
                va="center",
                color="#333",
            )
            # text.set_path_effects(
            #     [
            #         mpl.patheffects.Stroke(linewidth=2, foreground="#333333"),
            #         mpl.patheffects.Normal(),
            #     ]
            # )

        if symbol is not None:
            ax.annotate(
                text="$\\mathit{" + symbol + "}$\nexpression",
                xy=(1, len(cluster_info) / 2),
                xytext=(5, 0),
                textcoords="offset points",
                xycoords="axes fraction",
                ha="left",
                va="center",
            )

    @classmethod
    def from_transcriptome(
        cls, transcriptome, clustering, gene, width=0.5, panel_height=0.5, **kwargs
    ):
        transcriptome.adata.obs["cluster"] = clustering.labels
        plotdata_expression = sc.get.obs_df(
            transcriptome.adata, [gene, "cluster"]
        ).rename(columns={gene: "expression"})
        plotdata_expression_clusters = plotdata_expression.groupby("cluster")[
            "expression"
        ].mean()
        cluster_info = clustering.cluster_info

        return cls(
            plotdata_expression_clusters=plotdata_expression_clusters,
            cluster_info=cluster_info,
            width=width,
            panel_height=panel_height,
            **kwargs,
        )



import chromatinhd as chd
import numpy as np
import pandas as pd
import xarray as xr
import pickle
import scipy.stats
import tqdm.auto as tqdm
import torch


def fdr(p_vals):
    from scipy.stats import rankdata

    ranked_p_values = rankdata(p_vals)
    fdr = p_vals * len(p_vals) / ranked_p_values
    fdr[fdr > 1] = 1

    return fdr


class GenePositional(chd.flow.Flow):
    genes = chd.flow.Stored("genes", default=set)

    def score(
        self,
        fragments,
        clustering,
        models,
        folds,
        genes=None,
        force=False,
        device="cuda",
    ):
        force_ = force

        if genes is None:
            genes = fragments.var.index

        pbar = tqdm.tqdm(genes, leave=False)

        window = fragments.regions.window

        for gene in pbar:
            pbar.set_description(gene)
            probs_file = self.get_scoring_path(gene) / "probs.pkl"

            force = force_
            if not probs_file.exists():
                force = True

            if force:
                design_gene = pd.DataFrame(
                    {"gene_ix": [fragments.var.index.get_loc(gene)]}
                ).astype("category")
                design_gene.index = pd.Series([gene], name="gene")
                design_clustering = pd.DataFrame(
                    {"active_cluster": np.arange(clustering.n_clusters)}
                ).astype("category")
                design_clustering.index = clustering.cluster_info.index
                design_coord = pd.DataFrame(
                    {"coord": np.arange(window[0], window[1] + 1, step=25)}
                ).astype("category")
                design_coord.index = design_coord["coord"]
                design = chd.utils.crossing(
                    design_gene, design_clustering, design_coord
                )

                batch_size = 5000
                design["batch"] = np.floor(
                    np.arange(design.shape[0]) / batch_size
                ).astype(int)

                probs = []
                for model in models:
                    probs_model = []
                    for _, design_subset in design.groupby("batch"):
                        pseudocoordinates = torch.from_numpy(
                            design_subset["coord"].values.astype(int)
                        )
                        pseudocoordinates = (pseudocoordinates - window[0]) / (
                            window[1] - window[0]
                        )
                        pseudocluster = torch.nn.functional.one_hot(
                            torch.from_numpy(
                                design_subset["active_cluster"].values.astype(int)
                            ),
                            clustering.n_clusters,
                        ).to(torch.float)
                        gene_ix = torch.from_numpy(
                            design_subset["gene_ix"].values.astype(int)
                        )

                        prob = model.evaluate_pseudo(
                            pseudocoordinates,
                            clustering=pseudocluster,
                            gene_ix=gene_ix,
                        )

                        probs_model.append(prob.numpy())
                    probs_model = np.hstack(probs_model)
                    probs.append(probs_model)

                probs = np.vstack(probs)
                probs = probs.mean(axis=0)

                probs = xr.DataArray(
                    probs.reshape(  # we have only one gene anyway
                        (
                            design_clustering.shape[0],
                            design_coord.shape[0],
                        )
                    ),
                    coords=[
                        design_clustering.index,
                        design_coord.index,
                    ],
                )

                pickle.dump(probs, probs_file.open("wb"))

                self.genes = self.genes | {gene}

    def get_plotdata(self, gene):
        """
        Returns the plotdata for a given gene

        Parameters:
            gene:
                the gene

        Returns:
            Two dataframes, one with the probabilities per cluster, one with the mean
        """
        probs_file = self.get_scoring_path(gene) / "probs.pkl"
        if not probs_file.exists():
            raise FileNotFoundError(f"File {probs_file} does not exist")

        probs = pickle.load(probs_file.open("rb"))
        plotdata = probs.to_dataframe("prob")

        window = probs.coords["coord"].values[[0, -1]]

        plotdata["prob"] = (
            plotdata["prob"]
            - np.log(
                plotdata.reset_index()
                .groupby(["cluster"])
                .apply(
                    lambda x: np.trapz(
                        np.exp(x["prob"]),
                        x["coord"].astype(float) / (window[1] - window[0]),
                    )
                )
            ).mean()
        )
        plotdata_mean = plotdata[["prob"]].groupby("coord").mean()

        return plotdata, plotdata_mean

    def get_scoring_path(self, gene):
        path = self.path / f"{gene}"
        path.mkdir(parents=True, exist_ok=True)
        return path



import pandas as pd
from .dataset import Dataset


def get_canonical_transcripts(
    biomart_dataset: Dataset, gene_ids=None, chrom=None, start=None, end=None
):
    filters = []
    if gene_ids is not None:
        filters.append(
            biomart_dataset.filter("ensembl_gene_id", value=",".join(gene_ids))
        )
    if chrom is not None:
        filters.append(
            biomart_dataset.filter(
                "chromosome_name", value=str(chrom).replace("chr", "")
            )
        )
    if start is not None:
        filters.append(biomart_dataset.filter("start", value=str(start)))
    if end is not None:
        filters.append(biomart_dataset.filter("end", value=str(end)))
    genes = biomart_dataset.get(
        [
            biomart_dataset.attribute("ensembl_gene_id"),
            biomart_dataset.attribute("transcript_start"),
            biomart_dataset.attribute("transcript_end"),
            biomart_dataset.attribute("end_position"),
            biomart_dataset.attribute("start_position"),
            biomart_dataset.attribute("ensembl_transcript_id"),
            biomart_dataset.attribute("chromosome_name"),
            biomart_dataset.attribute("strand"),
            biomart_dataset.attribute("external_gene_name"),
            biomart_dataset.attribute("transcript_is_canonical"),
            biomart_dataset.attribute("transcript_biotype"),
        ],
        filters=filters,
    )
    genes["chrom"] = "chr" + genes["chromosome_name"].astype(str)
    genes = (
        genes.sort_values("transcript_is_canonical").groupby("ensembl_gene_id").first()
    )
    genes = genes.rename(
        columns={
            "transcript_start": "start",
            "transcript_end": "end",
            "external_gene_name": "symbol",
        }
    )
    genes.index.name = "gene"
    genes = genes.drop(
        [
            "chromosome_name",
            "transcript_biotype",
            "transcript_is_canonical",
            "start_position",
            "end_position",
        ],
        axis=1,
    )

    # remove genes without a symbol
    genes = genes.loc[~pd.isnull(genes["symbol"])]

    # remove genes on alternative contigs
    genes = genes.loc[~genes["chrom"].str.contains("_")]

    # remove genes on mitochondrial DNA
    genes = genes.loc[~genes["chrom"].isin(["chrMT", "chrM"])]

    # order according to gene_ids
    genes = genes.reindex(gene_ids)

    # add ensembl gene id
    genes["ensembl_gene_id"] = genes.index

    return genes


def get_exons(biomart_dataset: Dataset, chrom, start, end):
    canonical_transcripts = get_canonical_transcripts(
        biomart_dataset, chrom=chrom, start=start, end=end
    )
    exons = biomart_dataset.get(
        [
            biomart_dataset.attribute("ensembl_gene_id"),
            biomart_dataset.attribute("exon_chrom_start"),
            biomart_dataset.attribute("exon_chrom_end"),
            biomart_dataset.attribute("genomic_coding_start"),
            biomart_dataset.attribute("genomic_coding_end"),
            biomart_dataset.attribute("ensembl_transcript_id"),
        ],
        filters=[
            biomart_dataset.filter("start", value=str(start)),
            biomart_dataset.filter("end", value=str(end)),
            biomart_dataset.filter(
                "chromosome_name", value=str(chrom).replace("chr", "")
            ),
            biomart_dataset.filter(
                "ensembl_transcript_id",
                value=",".join(canonical_transcripts["ensembl_transcript_id"]),
            ),
        ],
    )

    return exons



import subprocess


class Colorcodes(object):
    """
    Provides ANSI terminal color codes which are gathered via the ``tput``
    utility. That way, they are portable. If there occurs any error with
    ``tput``, all codes are initialized as an empty string.
    The provides fields are listed below.
    Control:
    - bold
    - reset
    Colors:
    - blue
    - green
    - orange
    - red
    :license: MIT
    """

    def __init__(self):
        try:
            self.bold = subprocess.check_output("tput bold".split()).decode()
            self.reset = subprocess.check_output("tput sgr0".split()).decode()

            self.blue = subprocess.check_output("tput setaf 4".split()).decode()
            self.green = subprocess.check_output("tput setaf 2".split()).decode()
            self.orange = subprocess.check_output("tput setaf 3".split()).decode()
            self.red = subprocess.check_output("tput setaf 1".split()).decode()
        except subprocess.CalledProcessError as e:
            self.bold = ""
            self.reset = ""

            self.blue = ""
            self.green = ""
            self.orange = ""
            self.red = ""

    def color_sign(self, x, format):
        return (
            (self.red + format.format(x) + self.reset)
            if x >= 0
            else (self.green + format.format(x) + self.reset)
        )


colorcodes = Colorcodes()



from typing import Dict
import torch
import numpy as np
import scipy.sparse

from dataclasses import dataclass


def is_scipysparse(value):
    return isinstance(value, scipy.sparse.spmatrix)


def is_sparse(value):
    return isinstance(value, Sparse)

class Sparse:
    pass


@dataclass
class CSR(Sparse):
    """
    CSR sparse tensor
    We don't use the torch variant, because this is not supported on CUDA (and we would not use any of its functions anyway)
    """

    indptr: torch.tensor
    indices: torch.tensor
    values: torch.tensor
    shape: torch.Size

    @classmethod
    def from_scipy_csr(cls, scipy_csr, device="cpu", dtype=None):
        # will keep original numpy data type unless specified
        if dtype is None:
            values = torch.tensor(scipy_csr.data, device=device)
        else:
            values = torch.tensor(scipy_csr.data, device=device, dtype=dtype)
        return cls(
            indices=torch.tensor(scipy_csr.indices, device=device, dtype=torch.long),
            indptr=torch.tensor(scipy_csr.indptr, device=device, dtype=torch.long),
            values=values,
            shape=scipy_csr.shape,
        )


@dataclass
class COOMatrix(Sparse):
    """
    COOMatrix sparse matrix

    This implementation mainly cachges the row_switch and mapping dictionaries, which speed up subsampling in the first (row) dimension
    """

    row: torch.tensor
    col: torch.tensor
    values: torch.tensor
    shape: torch.Size
    row_switch = None
    mapping: Dict = None

    ndim = 2

    def populate_row_switch(self):
        device = self.row.device
        self.row_switch = torch.cat(
            (
                torch.tensor([0], device=device),
                torch.where(self.row[1:] - self.row[:-1])[0] + 1,
                torch.tensor([len(self.values)], device=device),
            )
        )
        assert (
            len(self.row_switch) == self.shape[0] + 1
        ), "There are probably some rows with only zeros, this is not supported"

    def populate_mapping(self):
        # this maps each row to its value and column indices
        # this allows us to go subset in the row dimension very quickly, by just extracting the relevant rows from this dictionary
        # at the cost of having to keep these mappings in memory
        device = self.row.device
        self.populate_row_switch()
        self.mapping = {
            row: torch.arange(
                self.row_switch[row], self.row_switch[row + 1], device=device
            )
            for row in range(self.shape[0])
        }

    def dense_slice(self, ix):
        """
        Slice (in the rows) and return a dense tensor. ix should be a slice.
        This is a bit faster, at the cost of requiring continuity in the subsampling
        Nonetheless, `dense_subset()` implementation is already quite fast, so this restriction is not necessary
        """
        return self.dense_subset(ix)

    def dense_subset(self, ix):
        """
        Subset (in the rows) and return a dense tensor. ix should be an np array.
        """

        if torch.is_tensor(ix):
            ix = ix.cpu().numpy()  #! This can be a major slow down
        elif isinstance(ix, slice):
            ix = np.arange(ix.start, ix.stop)
        elif isinstance(ix, np.ndarray):
            pass
        else:
            raise ValueError("Cannot take dense_subset of " + str(ix.__class__) + ", requires a tensor, slice or numpy.ndarray")

        device = self.row.device
        ix_sparse = torch.cat([self.mapping[row] for row in ix])
        newrow_idx = torch.arange(len(ix), device=device).repeat_interleave(
            self.row_switch[ix + 1] - self.row_switch[ix]
        )
        newcol_idx = self.col[ix_sparse]

        subsetted = torch.zeros(
            (len(ix), self.shape[1]), device=device, dtype=self.values.dtype
        )
        subsetted[newrow_idx, newcol_idx] = self.values[ix_sparse]
        return subsetted

    def dense(self):
        device = self.row.device
        subsetted = torch.zeros(self.shape, device=device, dtype=self.values.dtype)
        subsetted[self.row, self.col] = self.values
        return subsetted

    @classmethod
    def from_numpy_array(cls, ndarray, device="cpu", dtype=None):
        scipy_coo = scipy.sparse.coo_matrix(ndarray)
        return cls.from_scipy_coo(scipy_coo, device, dtype)

    @classmethod
    def from_scipy_coo(cls, scipy_coo, device="cpu", dtype=None):
        # will keep original numpy data type unless specified
        if dtype is None:
            values = torch.tensor(scipy_coo.data, device=device)
        else:
            values = torch.tensor(scipy_coo.data, device=device, dtype=dtype)
        return cls(
            row=torch.tensor(scipy_coo.row, device=device, dtype=torch.long),
            col=torch.tensor(scipy_coo.col, device=device, dtype=torch.long),
            values=values,
            shape=scipy_coo.shape,
        )

    @classmethod
    def from_scipy_csr(cls, scipy_csr, device="cpu", dtype=None):
        return cls.from_scipy_coo(scipy_csr.tocoo(), device=device, dtype=dtype)  # 😎

    @classmethod
    def from_scipy(cls, scipy_sparse, device="cpu", dtype=None):
        if isinstance(scipy_sparse, scipy.sparse.csr_matrix):
            return cls.from_scipy_csr(scipy_sparse, device=device, dtype=dtype)
        elif isinstance(scipy_sparse, scipy.sparse.coo_matrix):
            return cls.from_scipy_coo(scipy_sparse, device=device, dtype=dtype)
        else:
            raise NotImplementedError()

    def to_np(self):
        return self.dense().numpy()

    def to_scipy_coo(self):
        return scipy.sparse.coo_matrix(
            (self.values.numpy(), (self.row.numpy(), self.col.numpy())),
            shape=self.shape,
        )

    def to_scipy_csr(self):
        return self.to_scipy_coo().tocsr()

    def to_coo(self):
        return self

    def to(self, device):
        self.row = self.row.to(device)
        self.col = self.col.to(device)
        self.values = self.values.to(device)
        if self.row_switch is not None:
            self.row_switch = self.row_switch.to(device)
        if self.mapping is not None:
            self.mapping = {
                key: value.to(device) for key, value in self.mapping.items()
            }
        return self

    def __getitem__(self, ix):
        x = self.__class__.from_scipy_csr(self.to_scipy_csr()[:, ix[1]])
        x.populate_mapping()
        x.populate_row_switch()
        return x

    def __repr__(self):
        return f"{self.__class__}"



# ---
# jupyter:
#   jupytext:
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.14.7
#   kernelspec:
#     display_name: Python 3
#     language: python
#     name: python3
# ---

# %% [markdown]
# # ChromatinHD-*pred*

# %% tags=["hide_code", "hide_output"]
# autoreload
import IPython

if IPython.get_ipython() is not None:
    IPython.get_ipython().run_line_magic("load_ext", "autoreload")
    IPython.get_ipython().run_line_magic("autoreload", "2")

# %% tags=["hide_output"]
import chromatinhd as chd
import matplotlib.pyplot as plt

# %% [markdown]
# ChromatinHD-<i>pred</i> uses accessibility fragments to predict gene expression. As such, it can detect features such as broad or narrow positioning of fragments, or fragment sizes, that are predictive for gene expression.

# %% [markdown]
# We first load in all the input data:

# %%
import pathlib
dataset_folder = pathlib.Path("example")
fragments = chd.data.Fragments(dataset_folder / "fragments")
transcriptome = chd.data.Transcriptome(dataset_folder / "transcriptome")
folds = chd.data.folds.Folds(dataset_folder / "folds" / "5x1")

# %% [markdown]
# ## Train the models

# %% [markdown]
# The basic ChromatinHD-*pred* model

# %%
models = chd.models.pred.model.additive.Models(dataset_folder / "models" / "additive", reset = True)

# %% tags=["hide_output"]
models.train_models(fragments, transcriptome, folds, device = "cuda")

# %% [markdown]
# ## Some quality checks

# %% [markdown]
# We will first check whether the model learned something, by comparing the predictive performance with a baseline 

# %%
gene_cors = models.get_gene_cors(fragments, transcriptome, folds, device = "cuda")
gene_cors["symbol"] = gene_cors.index.map(transcriptome.symbol)

# %%
gene_cors.sort_values("deltacor", ascending = False).head(10)

# %%
import pandas as pd
import matplotlib.pyplot as plt

fig, ax = plt.subplots(figsize = (4, 4))

for name, group in gene_cors.iterrows():
    ax.plot([0, 1], group[["cor_n_fragments", "cor_predicted"]], color = "#3338", zorder = 0, marker = "o", markersize = 2)
ax.boxplot(gene_cors[["cor_n_fragments", "cor_predicted"]].values, positions = [0, 1], widths = 0.1, showfliers = False, showmeans = True, meanline = True, meanprops = {"color": "red", "linewidth": 2})
ax.set_xticks([0, 1])
ax.set_xticklabels(["# fragments", "ChromatinHD-pred"])
ax.set_ylabel("$cor$")
;

# %% [markdown]
# Note that every gene gains from the ChromatinHD model, even if some only gain a little. The genes with a low $\Delta cor$ are often those with only a few fragments:

# %%
fig, ax = plt.subplots(figsize = (4, 4))
ax.scatter(gene_cors["n_fragments"], gene_cors["deltacor"])
ax.set_ylabel("$\\Delta$ cor")
ax.set_xlabel("# fragments")
ax.set_xscale("log")

# %% [markdown]
# ## Predictivity per position

# %% [markdown]
# To determine which regions were important for the model to predict gene expression, we will censor fragments from windows of various sizes, and then check whether the model performance on a set of test cells decreased. This functionality is implemented in the `GeneMultiWindow` class. This will only run the censoring for a subset of genes to speed up interpretation.

# %%
censorer = chd.models.pred.interpret.MultiWindowCensorer(fragments.regions.window)
genemultiwindow = chd.models.pred.interpret.GeneMultiWindow(
    models.path / "interpret" / "genemultiwindow"
)

# %%
genemultiwindow.score(
    fragments,
    transcriptome,
    models,
    folds,
    transcriptome.gene_id(
        [
            "CCL4",
            "IL1B",
            "EBF1",
            "PAX5",
            "CD79A",
            "RHEX",
        ]
    ),
    censorer=censorer,
)

# %%
genemultiwindow.interpolate()

# %%
symbol = "EBF1"

fig = chd.grid.Figure(chd.grid.Grid(padding_height=0.05))
width = 10

region = fragments.regions.coordinates.loc[transcriptome.gene_id(symbol)]
panel_genes = chd.plot.genome.genes.Genes.from_region(region, width=width)
fig.main.add_under(panel_genes)

panel_pileup = chd.models.pred.plot.Pileup.from_genemultiwindow(
    genemultiwindow, transcriptome.gene_id(symbol), width=width
)
fig.main.add_under(panel_pileup)

panel_predictivity = chd.models.pred.plot.Predictivity.from_genemultiwindow(
    genemultiwindow, transcriptome.gene_id(symbol), width=width
)
fig.main.add_under(panel_predictivity)

fig.plot()

# %% [markdown]
# ## Co-predictivity per position

# %% [markdown]
# In a similar fashion we can determine the co-predictivity per position.

# %%
censorer = chd.models.pred.interpret.WindowCensorer(fragments.regions.window)
genepairwindow = chd.models.pred.interpret.GenePairWindow(
    models.path / "interpret" / "genepairwindow", reset = True
)
genepairwindow.score(fragments, transcriptome, models, folds, censorer = censorer, genes = transcriptome.gene_id(["CCL4"]))

# %%
symbol = "CCL4"

fig = chd.grid.Figure(chd.grid.Grid(padding_height=0.05))
width = 10

# genes
region = fragments.regions.coordinates.loc[transcriptome.gene_id(symbol)]
panel_genes = chd.plot.genome.genes.Genes.from_region(region, width=width)
fig.main.add_under(panel_genes)

# pileup
panel_pileup = chd.models.pred.plot.Pileup.from_genemultiwindow(
    genemultiwindow, transcriptome.gene_id(symbol), width=width
)
fig.main.add_under(panel_pileup)

# predictivity
panel_predictivity = chd.models.pred.plot.Predictivity.from_genemultiwindow(
    genemultiwindow, transcriptome.gene_id(symbol), width=width
)
fig.main.add_under(panel_predictivity)

# copredictivity
panel_copredictivity = chd.models.pred.plot.Copredictivity.from_genepairwindow(
    genepairwindow, transcriptome.gene_id(symbol), width=width
)
fig.main.add_under(panel_copredictivity)

fig.plot()

# %%
